# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Emulates various fp8 formats in bfloat16 or float32."""

from aqt.common import emulation_utils
import tensorflow as tf

from tensorflow.compiler.tf2xla.python import xla  # pylint: disable=g-direct-tensorflow-import

FPMetadata = emulation_utils.FPMetadata
EXPONENT_BIAS = emulation_utils.EXPONENT_BIAS
get_max_number = emulation_utils.get_max_number



def get_exponent(t):
  """Get exponent of a tensor."""

  assert t.dtype in [tf.bfloat16, tf.float32]

  t_shape = t.shape

  # Make MSB zero
  t = tf.abs(t)

  if t.dtype == tf.bfloat16:
    t = tf.bitcast(t, tf.int16)
    t = tf.bitwise.right_shift(t, tf.constant(7, dtype=tf.int16))
    assert t.dtype == tf.int16
  else:
    t = tf.bitcast(t, tf.int32)
    t = tf.bitwise.right_shift(t, tf.constant(23, dtype=tf.int32))
    assert t.dtype == tf.int32

  assert t.shape == t_shape
  return t - EXPONENT_BIAS


def flush_to_zero(t, exponent, min_exp):
  """Handles underflow when the exponent is less than min_exp.

  Currently flush all underflow to zeros and does not handle subnormals.

  Args:
    t: a tensor whose dtype is either tf.bfloat16 or tf.float32
    exponent: the exponents of t.
    min_exp: the allowed minimum exponents.

  Returns:
    a tensor in the same dtype with underflow flushed to zeros.
  """

  assert t.shape == exponent.shape
  assert t.dtype in [tf.bfloat16, tf.float32]

  if t.dtype == tf.bfloat16:
    int_dtype = tf.int16
  else:
    int_dtype = tf.int32

  assert exponent.dtype == int_dtype
  if isinstance(min_exp, tf.Tensor):
    min_exp = tf.cast(min_exp, dtype=int_dtype) + EXPONENT_BIAS
  else:
    min_exp = tf.constant(min_exp, dtype=int_dtype) + EXPONENT_BIAS
  zeros = tf.zeros_like(t, dtype=t.dtype)
  mask = tf.greater_equal(exponent + EXPONENT_BIAS, min_exp)
  # The following might change the sign of zeros.
  return tf.where(mask, t, zeros)


def get_max_frac_in_binary(mantissa_bits, dtype):
  """Returns the maximum fraction in binary format (dtype-aware)."""
  if dtype == tf.float32:
    if mantissa_bits == 2:
      return 0x00600000
    else:
      return 0x00700000
  elif dtype == tf.bfloat16:
    if mantissa_bits == 2:
      return 0x0060
    else:
      return 0x0070


def handle_overflow(t, exponent, max_exp, mantissa_bits):
  """Handles exponent overflow of the input tensor.

  Since the definition of NaN/inf is format dependent. This function assumes
  no such special values (NaN or inf).

  Args:
    t: a tensor whose dtype is either tf.bfloat16 or tf.float32
    exponent: the exponents of t.
    max_exp: the allowed maximum exponents.
    mantissa_bits: the number of mantissa bits in the format.

  Returns:
    a tensor in the same dtype with overflow clamped at the max possible value.
  """

  assert t.shape == exponent.shape
  assert t.dtype in [tf.bfloat16, tf.float32]

  if t.dtype == tf.bfloat16:
    int_dtype = tf.int16
  else:
    int_dtype = tf.int32

  if isinstance(max_exp, tf.Tensor):
    max_exp = tf.cast(max_exp, dtype=int_dtype) + EXPONENT_BIAS
  else:
    max_exp = tf.constant(max_exp, dtype=int_dtype) + EXPONENT_BIAS

  assert exponent.dtype == int_dtype
  max_frac = get_max_frac_in_binary(mantissa_bits, t.dtype)
  if t.dtype == tf.bfloat16:
    abs_max_value = tf.bitwise.left_shift(max_exp,
                                          tf.constant(7, dtype=tf.int16))
    abs_max_value = tf.bitwise.bitwise_or(abs_max_value, max_frac)
    abs_max_value = tf.bitcast(abs_max_value, tf.bfloat16)
  else:
    abs_max_value = tf.bitwise.left_shift(max_exp,
                                          tf.constant(23, dtype=tf.int32))
    abs_max_value = tf.bitwise.bitwise_or(abs_max_value, max_frac)
    abs_max_value = tf.bitcast(abs_max_value, tf.float32)

  return tf.clip_by_value(
      t, clip_value_min=-abs_max_value, clip_value_max=abs_max_value)


def get_mask(dtype, num_lsbs):
  """Get mask with number of LSBs set to 0."""

  assert dtype in [tf.uint16, tf.uint32]
  assert isinstance(num_lsbs, tf.Tensor)

  if dtype == tf.uint16:
    mask = 0xFFFF
  else:
    mask = 0xFFFFFFFF

  mask -= tf.math.pow(2, tf.cast(num_lsbs, tf.int32)) - 1
  mask = tf.cast(mask, dtype)
  return mask


def _round_stochastic(t, mask_bits):
  """Round to a near value with a probability dependent on the proximity.

  Implementations-wise, add a rounding bias that is a random number in [0, 1),
  followed by a truncation. The random number is based on the number of mask
  bits. For example, converting bf16 (E8M7) to E5M2, the number of mask bits is
  5. The random value is a relative term between [0, 1.0) that are
  representable by 5 mantissa bits. That is, the rounding bias is a uniform
  distribution in [0.0000000, 0.0100000) for the mantissa bits. See the below
  example for more details.

  Use the following example, the values are in the format of 1.x * 2^y.
  exponent: y=4 (2^4=16)
  mantissa: 2 bits (in binary): 1.00, 1.01, 1.10, and 1.11.
  representable values: 16, 20, 24, 28. The next representable value is 32
  with an exponent of 4 and a mantissa of 0 (1.00 * 2^5 = 32).

  Take a value of 17 as an example.  The rounding bias is
  [0.0000000, 0.0100000) * 2^4, which is [0, 4). The result of the addition is
  [17, 21), and the following truncation produces a mix of 75% of 16 and 25% of
  20.

  For a negative value of -17, the rounding bias is the same but used in
  subtraction. The result of the subtraction is (-21, -17], and the
  following truncation produces a mix of 75% of -16 and 25% of -20.

  https://en.wikipedia.org/wiki/Rounding#Stochastic_rounding

  Args:
    t: the tensor in integer dtype.
    mask_bits: the number of mask bits.

  Returns:
    a tensor that is rounded with stochastic rounding.
  """
  t_shape = t.shape
  int_dtype = t.dtype

  # add rounding bias.
  if t.dtype is tf.uint16:
    upper = tf.math.pow(2, 7)
  elif t.dtype is tf.uint32:
    upper = tf.math.pow(2, 23)
  rounding_bias = tf.cast(
      tf.random.uniform(
          t_shape,
          minval=0,
          maxval=upper,
          dtype=tf.dtypes.int32,
          seed=0,
      ), int_dtype)
  # Use mask to deal with subnormals.
  rounding_bias_mask = tf.bitwise.left_shift(
      tf.constant(1, int_dtype), mask_bits) - 1
  rounding_bias = tf.bitwise.bitwise_and(rounding_bias, rounding_bias_mask)
  t = t + rounding_bias

  # Truncation
  mask = get_mask(int_dtype, mask_bits)
  t = tf.bitwise.bitwise_and(t, mask)
  assert t.shape == t_shape

  return t


def _round_to_nearest_even(t, mask_bits):
  """Round to nearest even value.

  Add a rounding bias for positive, subtract the bias for negative. The rounding
  bias is depending on whether the value is odd or even. For odd values, the
  addition is "half", which depends on the number of mantissa bits. For 2 bits,
  the "half" value is 0.125 (the third mantissa bit). If the mantissa overflows,
  the mechanism is to increment the exponent. The following is an example. For
  even values, the addition is the value that is smaller than but closest to 0.5
  (for example, 0.4999).

  Use the following example, the values are in the format of 1.x * 2^y.
  exponent: y=3 (2^3=8)
  mantissa: 2 bits: 1.0, 1.25, 1.5, and 1.75. Among these four values, 1.0
            (0x00) and 1.5 (0x10) are considered even values.
  representable values: 8, 10, 12, 14. The next representable value is 16
  with an exponent of 4 and a mantissa of 0. Among these values, 8, 12, and 16
  are even values.
  values in [8, 9] become 8 (round-to-nearest-even).
  values in (9, 11) become 10.
  values in [11, 13] becomes 12
  values in (13, 15) become 14.
  values in [15, 16] become 16 (increment on the exponent).

  https://en.wikipedia.org/wiki/Rounding#Round_half_to_even

  Args:
    t: the tensor in integer dtype.
    mask_bits: the number of mask bits.

  Returns:
    a tensor that is rounded with round-to-nearest-even.
  """
  t_shape = t.shape
  int_dtype = t.dtype

  # add rounding bias.
  rounding_bias = tf.bitwise.left_shift(
      tf.constant(1, int_dtype), mask_bits - 1) - 1
  even_mask = tf.bitwise.left_shift(tf.constant(1, int_dtype), mask_bits)
  is_even = tf.bitwise.bitwise_and(t, even_mask)
  # The rounding bias depends on whether the value is odd or even.
  rounding_bias = tf.where(
      tf.equal(is_even, tf.zeros_like(is_even)), rounding_bias,
      rounding_bias + 1)

  t = t + rounding_bias

  # Truncation
  mask = get_mask(int_dtype, mask_bits)
  t = tf.bitwise.bitwise_and(t, mask)
  assert t.shape == t_shape

  return t


def _round_away_from_zero(t, mask_bits):
  """Round to nearest value with ties away from zero.

  Add half for positive, subtract half for negative. The value of "half"
  depends on the number of mantissa bits. For 2 bits, the "half" value is
  0.125 (the third mantissa bit). If the mantissa overflows, the mechanism is
  to increment the exponent. The following is an example.
  exponent: 3 (2^3)
  mantissa: 2 bits.
  representable values: 8, 10, 12, 14. The next representable value is 16
  with an exponent of 4 and a mantissa of 0.
  values in [8, 9) become 8.
  values in [9, 10] become 10. (round-half-away-from-zero).
  values in [14, 15) become 14.
  values in [15, 16] become 16 (increment on the exponent).

  https://en.wikipedia.org/wiki/Rounding#Round_half_away_from_zero

  Args:
    t: the tensor in integer dtype.
    mask_bits: the number of mask bits.

  Returns:
    a tensor that is rounded with round-half-away-from-zero.
  """
  t_shape = t.shape
  int_dtype = t.dtype

  # add half
  half = tf.bitwise.left_shift(tf.constant(1, int_dtype), mask_bits - 1)
  t = t + half

  # Truncation
  mask = get_mask(int_dtype, mask_bits)
  t = tf.bitwise.bitwise_and(t, mask)
  assert t.shape == t_shape

  return t


def rounding(t, mask_bits, rounding_mode):
  """Rounding given the mask bits."""
  t_shape = t.shape
  dtype = t.dtype

  if t.dtype == tf.bfloat16:
    int_dtype = tf.uint16
  else:
    int_dtype = tf.uint32
  # Use a integer dtype as a container.
  t = tf.bitcast(t, int_dtype)
  assert t.shape == t_shape

  if rounding_mode == emulation_utils.ROUND_AWAY_FROM_ZERO:
    t = _round_away_from_zero(t, mask_bits)
  elif rounding_mode == emulation_utils.ROUND_TO_NEAREST_EVEN:
    t = _round_to_nearest_even(t, mask_bits)
  elif rounding_mode == emulation_utils.ROUND_STOCHASTIC:
    t = _round_stochastic(t, mask_bits)
  else:
    raise NotImplementedError

  # convert to bfloat16/float32
  t = tf.bitcast(t, dtype)
  assert t.shape == t_shape
  return t


def handle_mantissa(t, mantissa_bits, min_exp, rounding_mode):
  """Handles the mantissa of t with the given mantissa bits.

  This function handles the mantissa of t by the given mantissa bits with
  the support of subnormals.

  Args:
    t: a tensor whose dtype is either tf.bfloat16 or tf.float32
    mantissa_bits: the number of allowed mantissa bits.
    min_exp: an integer number for the minimum exponent.
    rounding_mode: the rounding mode.

  Returns:
    a tensor in the same dtype with manipulated mantissa.
  """

  assert t.dtype in [tf.bfloat16, tf.float32]
  if t.dtype == tf.bfloat16:
    int_dtype = tf.uint16
  else:
    int_dtype = tf.uint32

  exponent = get_exponent(t)
  allowed_mantissa_bits = tf.cast(
      tf.subtract(
          tf.cast(mantissa_bits, exponent.dtype),
          tf.clip_by_value(
              tf.subtract(tf.cast(min_exp, exponent.dtype), exponent), 0,
              mantissa_bits)),
      dtype=int_dtype)

  if t.dtype is tf.bfloat16:
    mask_bits = 7 - allowed_mantissa_bits
  elif t.dtype is tf.float32:
    mask_bits = 23 - allowed_mantissa_bits

  return rounding(t, mask_bits, rounding_mode)


def static_handle_exponent(t, min_exp, max_exp, mantissa_bits):
  """Handles the exponents of inputs by a limiting range [min_exp, max_exp].

  This function limits the range of exponents of inputs in a static way.

  Args:
    t: a tensor whose dtype is either tf.bfloat16 or tf.float32
    min_exp: the allowed minimum exponents.
    max_exp: the allowed maximum exponents.
    mantissa_bits: the number of mantissa bits in the format.

  Returns:
    a tensor in the same dtype with a limited range of exponents.
  """
  exponent = get_exponent(t)
  value = flush_to_zero(t, exponent, min_exp)
  value = handle_overflow(value, exponent, max_exp, mantissa_bits)
  return value


# Emulated fp8 subnormal handling
# Using E5M2B15 as an example. E5M2B15: 1 sign bit, 5 exponent bits,
# 2 mantissa bits, and the bias is 15.
# The smallest positive normalized number is 2 ** -14: exponent is 1 and
# mantissa are all zeros. That is, 0x04. For subnormals, the exponent is zero,
# and the mantissa is non-zero. This emulated version handles subnormals by
# masking away more and more low order bits as the f32 exponent falls.
# e4m3 variants have different exponent, mantissa, and bias. Therefore, the
# representable subnormals differ.
def emulated_fp(t, fp_metadata):
  """Emulates enumerics in bfloat16 or float32 given the FPMetadata.

  While zeros, NaN, and subnormals are supported, the support of inf is
  specified by the FPMetadata. This function is generic to support a variety
  of FPn formats.

  Args:
    t: a tensor whose dtype is either tf.bfloat16 or tf.float32.
    fp_metadata: a FPMetadata object.

  Returns:
    a same dtype tensor that emulates floating points.
  """
  assert isinstance(fp_metadata, FPMetadata)
  assert t.dtype in [tf.bfloat16, tf.float32]

  with tf.name_scope('emulated_fp'):
    v = handle_mantissa(
        t,
        mantissa_bits=fp_metadata.mantissa_bits,
        min_exp=fp_metadata.min_exp,
        rounding_mode=fp_metadata.rounding_mode)
    v = static_handle_exponent(
        v,
        min_exp=fp_metadata.min_exp - fp_metadata.mantissa_bits,
        max_exp=fp_metadata.max_exp,
        mantissa_bits=fp_metadata.mantissa_bits)

    # Supporting inf is optional in some of the formats. If inf is supported,
    # the emulation preserves nan and inf and changes the input whose value is a
    # finite number. Otherwise, the emulation changes the input whose value is
    # not nan (and treat inf as a valid number.)
    def cond(t):
      if fp_metadata.support_inf:
        return tf.math.is_finite(t)
      else:
        return tf.math.logical_not(tf.math.is_nan(t))

    return tf.where(cond(t), v, t)


# Emulated E8Mn, n being 1 to 6.
# Use xla reduce precision ops for the emulation.
def emulated_e8mn(t, fp_metadata):
  """Emulates E8Mn in bfloat16 or float32 given the FPMetadata.

  Special value support depends on the underlying hardware.

  Args:
    t: a tensor whose dtype is either tf.bfloat16 or tf.float32.
    fp_metadata: a FPMetadata object.

  Returns:
    a same dtype tensor that emulates floating points.
  """
  assert isinstance(fp_metadata, FPMetadata)
  assert fp_metadata.exponent_bits == 8
  assert fp_metadata.mantissa_bits <= 6 and fp_metadata.mantissa_bits >= 0
  assert t.dtype in [tf.bfloat16, tf.float32]

  with tf.name_scope('emulated_e8mn'):
    return xla.reduce_precision(t, fp_metadata.exponent_bits,
                                fp_metadata.mantissa_bits)




# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import operator

from absl.testing import parameterized
from aqt.jax.v2 import aqt_quantizer
from aqt.jax.v2 import calibration
from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import int_numerics
import jax.numpy as jnp
import numpy as np

from google3.testing.pybase import googletest


def inclusive_arange(start, stop, step):
  return jnp.arange(start, stop + step, step)


def assert_array_less_or_equal(x, y, err_msg='', verbose=True):
  np.testing.assert_array_compare(
      operator.__le__,
      x,
      y,
      err_msg=err_msg,
      verbose=verbose,
      header='x is not less than or equal to y',
      equal_inf=False,
  )


def assert_array_greater_or_equal(x, y, err_msg='', verbose=True):
  np.testing.assert_array_compare(
      operator.__ge__,
      x,
      y,
      err_msg=err_msg,
      verbose=verbose,
      header='x is not greater than or equal to y',
      equal_inf=False,
  )


class IntSymmetricTest(parameterized.TestCase):

  @parameterized.product(
      bits=[2, 4],
      preserve_zero=[False, True],
      preserve_max_val=[False, True],
  )
  def test_quant_range(
      self,
      bits,
      preserve_zero,
      preserve_max_val,
  ):
    sint_min = -(2 ** (bits - 1))
    sint_max = 2 ** (bits - 1) - 1
    sint_min_restricted = sint_min + 1

    def quantize(x):
      numerics_ = int_numerics.IntSymmetric(
          bits=bits,
          preserve_zero=preserve_zero,
          preserve_max_val=preserve_max_val,
          # The quantized values are only guarenteed to be within the
          # appropriate signed int range if clip=True and round=True.
          clip=True,
          clip_gradient=False,
          round=True,
          noise_fn=None,
          dtype=jnp.int8,
      )
      q = aqt_quantizer.Quantizer(
          numerics=numerics_,
          calib_shared_axes=None,
          scale_stop_grad=True,
          calibration=calibration.AbsMaxCalibration,
          context=utils.Context(key=None, train_step=None),
      )
      q.init_calibration()
      qx, _ = q.quant(x, calibration_axes=-1)
      return qx

    step_size = 0.25
    qx_neg_max_abs = quantize(inclusive_arange(-2, 1, step_size))
    qx_eq_max_abs = quantize(inclusive_arange(-1, 1, step_size))
    qx_pos_max_abs = quantize(inclusive_arange(-1, 2, step_size))

    # Test that the quantized value is not outside of the signed int range.
    # E.g. [-2, 1] for bits=2.
    assert_array_less_or_equal(qx_neg_max_abs.qvalue, sint_max)
    assert_array_less_or_equal(qx_eq_max_abs.qvalue, sint_max)
    assert_array_less_or_equal(qx_pos_max_abs.qvalue, sint_max)
    assert_array_greater_or_equal(qx_neg_max_abs.qvalue, sint_min)
    assert_array_greater_or_equal(qx_eq_max_abs.qvalue, sint_min)
    assert_array_greater_or_equal(qx_pos_max_abs.qvalue, sint_min)

    # Test that the quantized value is in the restricted symmetric range.
    # E.g. -127, 127 for bits=8 and [-1, 1] for bits=2.
    assert_array_greater_or_equal(qx_neg_max_abs.qvalue, sint_min_restricted)
    assert_array_greater_or_equal(qx_eq_max_abs.qvalue, sint_min_restricted)
    assert_array_greater_or_equal(qx_pos_max_abs.qvalue, sint_min_restricted)


if __name__ == '__main__':
  googletest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for flax e2e model."""
import copy
from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import config
from aqt.jax.v2 import tiled_dot_general
from aqt.jax.v2.examples import flax_e2e_model
from aqt.jax.v2.flax.intercept import aqt_intercept_methods
from aqt.jax.v2.flax.intercept.examples import flax_e2e_intercept_model
import jax
import jax.numpy as jnp


class MnistTest(parameterized.TestCase):

  # Unable to use config_v4() in parameters since it needs jax.device info.
  # TODO(aqt): Move confiv_v4() into parameters once int4 works for cpu.
  @parameterized.parameters([
      (
          {
              "drhs_bits": 8,
              "drhs_accumulator_dtype": jnp.int32,  # overwrite the default None
          },
          8,
      ),
      (
          {
              "fwd_bits": 4,
              "fwd_accumulator_dtype": None,
              "dlhs_accumulator_dtype": None,
          },
          4,
      ),
  ])
  def test_mnist_training(self, configs, bits):
    aqt_cfg = config.config_v4(**configs)
    tiling_cfg = tiled_dot_general.Cfg(
        lhs=tiled_dot_general.TensorTiling(
            contraction_axes=[
                tiled_dot_general.AxisTiling(
                    axis=1, tile_count=2, tile_size=None
                ),
            ],
            remaining_axes=[],
        ),
        rhs=tiled_dot_general.TensorTiling(
            contraction_axes=[
                tiled_dot_general.AxisTiling(
                    axis=0, tile_count=2, tile_size=None
                ),
            ],
            remaining_axes=[],
        ),
    )
    target_loss = {
        8: {
            "TPU v5 lite": [3.222228527069091796875000000000],
        },
        4: {
            "TPU v5 lite": [2.292296886444091796875000000000],
        },
    }
    # below 3 lines are differences between config_v4/v3 and fully_quantized
    config.set_stochastic_rounding(aqt_cfg, True, True, "jax.uniform")
    aqt_cfg.dlhs.rhs.use_fwd_quant = True
    aqt_cfg.drhs.rhs.use_fwd_quant = True

    def forward(model, apply_fn):
      return apply_fn(
          model,
          ds["image"],
          rngs={"params": jax.random.PRNGKey(0)},
          mutable=True,
      )

    # RNGs
    rng = jax.random.key(0)
    rng, init_rng = jax.random.split(rng)
    rng, image_rng = jax.random.split(rng)
    rng, label_rng = jax.random.split(rng)
    rng, input_rng = jax.random.split(rng)
    del rng

    # Dataset
    ds_size = 8
    ds = {
        "image": jax.random.uniform(key=image_rng, shape=(ds_size, 28, 28, 1)),
        "label": jax.random.randint(
            key=label_rng, shape=(ds_size,), minval=0, maxval=10
        ),
    }

    # Stage 1: regular training
    state = flax_e2e_intercept_model.create_train_state(init_rng)

    with aqt_intercept_methods.intercept_methods(
        aqt_cfg, tiling_cfg=tiling_cfg):
      state, train_loss, _ = flax_e2e_model.train_epoch(
          state, ds, batch_size=ds_size // 2, rng=input_rng
      )

    device_kind = jax.devices()[0].device_kind
    expected_train_loss = target_loss[bits][device_kind]
    if train_loss not in expected_train_loss:
      msg = "train_loss changed. Consider updating with the following:\n"
      msg += f'        "{device_kind}": [{train_loss:.30f}]'
      self.fail(msg)

    # Run forward once more in the same mode to get logits for testing below.
    with aqt_intercept_methods.intercept_methods(
        aqt_cfg, tiling_cfg=tiling_cfg):
      logits_s1, _ = forward(state.model, state.cnn_eval.apply)

    # Stage 2: Model conversion (quantized weights freezing)

    apply_serving, model_serving = flax_e2e_intercept_model.serving_conversion(
        state, aqt_cfg, tiling_cfg)

    dtype = jnp.dtype
    expected_dtype = jnp.int4 if bits == 4 else jnp.int8
    expected_aqt_pytree = {
        "aqt": {
            "Dense_0": {
                "AqtDotGeneral_0": {
                    "qrhs": {
                        # The scale_t shape was (1, 256) before tiling.
                        # After tiling the scale shape is (2, 1, 256),
                        # then transposed to (2, 1, 256).
                        "scale": (dtype("float32"), (2, 1, 256)),
                        # The weight shape was (3136, 256) before tiling.
                        # After tiling it is (2, 1568, 256).
                        # Contraction shape 3136 is tiled to (2, 1568).
                        # The remaining shape 256 is not tiled.
                        "value": (expected_dtype, (2, 1568, 256)),
                    }
                }
            },
            "Dense_1": {
                "AqtDotGeneral_0": {
                    "qrhs": {
                        # The scale_t shape was (1, 10) before tiling.
                        # After tiling the scale shape is (2, 1, 10),
                        # then transposed to (2, 1, 10).
                        "scale": (dtype("float32"), (2, 1, 10)),
                        # The weight shape was (256, 10) before tiling.
                        # After tiling it is (2, 128, 10).
                        # Contraction shape 256 is tiled to (2, 128).
                        # The remaining shape 10 is not tiled.
                        "value": (expected_dtype, (2, 128, 10)),
                    }
                }
            },
        },
        "batch_stats": {
            "BatchNorm_0": {
                "mean": (dtype("float32"), (32,)),
                "var": (dtype("float32"), (32,)),
            },
            "BatchNorm_1": {
                "mean": (dtype("float32"), (64,)),
                "var": (dtype("float32"), (64,)),
            },
        },
        "params": {
            "BatchNorm_0": {
                "bias": (dtype("float32"), (32,)),
                "scale": (dtype("float32"), (32,)),
            },
            "BatchNorm_1": {
                "bias": (dtype("float32"), (64,)),
                "scale": (dtype("float32"), (64,)),
            },
            "Conv_0": {
                "bias": (dtype("float32"), (32,)),
                "kernel": (dtype("float32"), (3, 3, 1, 32)),
            },
            "Conv_1": {
                "bias": (dtype("float32"), (64,)),
                "kernel": (dtype("float32"), (3, 3, 32, 64)),
            },
            "Dense_0": {
                "bias": (dtype("float32"), (256,)),
                "kernel": (dtype("float32"), (3136, 256)),
            },
            "Dense_1": {
                "bias": (dtype("float32"), (10,)),
                "kernel": (dtype("float32"), (256, 10)),
            },
        },
    }

    serving_pytree = jax.tree_util.tree_map(
        lambda x: (x.dtype, x.shape), model_serving
    )
    if serving_pytree != expected_aqt_pytree:
      print()
      print("serving_pytree:      ", serving_pytree)
      print("expected_aqt_pytree: ", expected_aqt_pytree)
      assert False, "serving_pytree != expected_aqt_pytree"

    def zero_out_params(model, layer: str):
      updated_model = copy.deepcopy(model)
      updated_model["params"][layer]["kernel"] = jnp.zeros_like(
          updated_model["params"][layer]["kernel"]
      )
      return updated_model

    # We can, but do not have to delete unquantized weights.
    model_serving = zero_out_params(model_serving, "Dense_0")
    model_serving = zero_out_params(model_serving, "Dense_1")

    # Stage 3: inference mode.
    logits_s3, _ = forward(model_serving, apply_serving)
    assert (logits_s3 == logits_s1).all()

    # Sanity check 1: We can't zero out Conv_0 because it was not frozen.
    sanity_model = zero_out_params(model_serving, "Conv_0")
    bad_logits, _ = forward(sanity_model, apply_serving)
    assert not (bad_logits == logits_s3).all()

    # Sanity check 2: Frozen weights are indeed used for inference.
    #   If we zero them out, loss would change.
    model_serving["aqt"]["Dense_0"]["AqtDotGeneral_0"]["qrhs"]["value"] = (
        jnp.zeros_like(
            model_serving["aqt"]["Dense_0"]["AqtDotGeneral_0"]["qrhs"]["value"]
        )
    )
    bad_logits, _ = forward(model_serving, apply_serving)
    assert not (bad_logits == logits_s3).all()


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for AQT Tiled Dot General."""
import functools
import itertools

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import tiled_dot_general
from aqt.jax.v2 import utils
import jax
import jax.numpy as jnp
import numpy as np
import sympy


AxisIdx = utils.AxisIdx
TensorTiling = tiled_dot_general.TensorTiling
Cfg = tiled_dot_general.Cfg
AxisTiling = tiled_dot_general.AxisTiling


def get_shape_from_axes(axes: list[AxisIdx], shape) -> list[int]:
  return [shape[i] for i in axes]


def assign_input_shape(
    rng_key, ca_shape: list[int], ba_shape: list[int], ra_shape: list[int]
):
  """Randomly assign an axis to a shape."""
  num_ca = len(ca_shape)
  num_ba = len(ba_shape)
  num_ra = len(ra_shape)
  total_num_axes = num_ca + num_ba + num_ra
  axes = jax.random.permutation(rng_key, total_num_axes).tolist()
  ca = axes[:num_ca]
  ba = axes[num_ca : num_ca + num_ba]
  ra = axes[num_ca + num_ba :]
  raw_input_shape = np.array(ca_shape + ba_shape + ra_shape)
  input_shape = raw_input_shape[np.argsort(axes)].tolist()
  return ca, ba, ra, input_shape


def get_axis_tiles(
    axes: list[AxisIdx], input_shape: list[int], subsample_ts: int = 2
) -> list[list[AxisTiling]]:
  comb = [list(itertools.combinations(axes, i)) for i in range(len(axes) + 1)]
  axes_tile_choices = list(itertools.chain(*comb))  # flatten lists
  ret = []
  for axes_to_tile in axes_tile_choices:
    selected_axis_cfg = []
    for axis in axes_to_tile:
      # Only consider testing part of the tile sizes as they have little impact.
      all_ts = sympy.divisors(input_shape[axis])[::subsample_ts]
      selected_axis_cfg.append(
          [tiled_dot_general.AxisTiling(axis, None, ts) for ts in all_ts]
      )
    tile_cfg = list(itertools.product(*selected_axis_cfg))
    ret.extend(list(tile_cfg))

  return ret


def generate_inputs(
    rng_key,
    num_ca: int,
    num_ba: int,
    num_lhs_ra: int,
    num_rhs_ra: int,
    max_shape_val: int,
):
  keys = jax.random.split(rng_key, 6)

  ca_shape = jax.random.randint(
      keys[0], shape=(num_ca,), minval=1, maxval=max_shape_val
  ).tolist()
  ba_shape = jax.random.randint(
      keys[1], shape=(num_ba,), minval=1, maxval=max_shape_val
  ).tolist()
  lhs_ra_shape = jax.random.randint(
      keys[2], shape=(num_lhs_ra,), minval=1, maxval=max_shape_val
  ).tolist()
  rhs_ra_shape = jax.random.randint(
      keys[3], shape=(num_rhs_ra,), minval=1, maxval=max_shape_val
  ).tolist()

  lhs_ca, lhs_ba, lhs_ra, lhs_shape = assign_input_shape(
      keys[4], ca_shape, ba_shape, lhs_ra_shape
  )
  rhs_ca, rhs_ba, rhs_ra, rhs_shape = assign_input_shape(
      keys[5], ca_shape, ba_shape, rhs_ra_shape
  )
  assert get_shape_from_axes(lhs_ca, lhs_shape) == get_shape_from_axes(
      rhs_ca, rhs_shape
  ), 'Found lhs ca and rhs ca shapes are different.'
  assert get_shape_from_axes(lhs_ba, lhs_shape) == get_shape_from_axes(
      rhs_ba, rhs_shape
  ), 'Found lhs ba and rhs ba shapes are different.'

  lhs = jnp.ones(lhs_shape)
  rhs = jnp.ones(rhs_shape)
  return lhs, rhs, (lhs_ca, lhs_ba, lhs_ra), (rhs_ca, rhs_ba, rhs_ra)


def generate_tiling_cfgs(lhs_shape, lhs_ca, lhs_ra, rhs_shape, rhs_ca, rhs_ra):
  lhs_remaining_axis_tilings = get_axis_tiles(lhs_ra, lhs_shape)
  lhs_contraction_axes_tilings = get_axis_tiles(lhs_ca, lhs_shape)
  rhs_remaining_axis_tilings = get_axis_tiles(rhs_ra, rhs_shape)
  rhs_contraction_axes_tilings = get_axis_tiles(rhs_ca, rhs_shape)

  cfgs = []
  assert len(lhs_contraction_axes_tilings) == len(rhs_contraction_axes_tilings)
  for i in range(len(lhs_contraction_axes_tilings)):
    for j in range(len(lhs_remaining_axis_tilings)):
      for k in range(len(rhs_remaining_axis_tilings)):
        cfgs.append(
            tiled_dot_general.Cfg(
                lhs=tiled_dot_general.TensorTiling(
                    contraction_axes=list(lhs_contraction_axes_tilings[i]),
                    remaining_axes=list(lhs_remaining_axis_tilings[j]),
                ),
                rhs=tiled_dot_general.TensorTiling(
                    contraction_axes=list(rhs_contraction_axes_tilings[i]),
                    remaining_axes=list(rhs_remaining_axis_tilings[k]),
                ),
            )
        )
  print(f'{len(cfgs)} configs in total.')
  return cfgs


class TiledDotGeneralTest(parameterized.TestCase):

  def test_tiled_dot_general_shape(self):
    for i in range(10):
      key1, key2 = jax.random.split(jax.random.PRNGKey(i), 2)
      hypers = jax.random.randint(key1, shape=(4,), minval=2, maxval=5)
      num_ca = hypers[0]
      num_ba = hypers[1]
      num_lhs_ra = hypers[2]
      num_rhs_ra = hypers[3]
      max_shape_val = 6
      lhs, rhs, lhs_axes, rhs_axes = generate_inputs(
          key2, num_ca, num_ba, num_lhs_ra, num_rhs_ra, max_shape_val
      )
      lhs_ca, lhs_ba, lhs_ra = lhs_axes
      rhs_ca, rhs_ba, rhs_ra = rhs_axes
      dims = ((tuple(lhs_ca), tuple(rhs_ca)), (tuple(lhs_ba), tuple(rhs_ba)))

      def _lax_dg(dims, lhs_in, rhs_in):
        return jax.lax.dot_general(lhs_in, rhs_in, dims)

      def _tiled_dg(cfg_in, dims, lhs_in, rhs_in):
        return tiled_dot_general.tiled_dot_general(cfg_in, lhs_in, rhs_in, dims)

      lax_dg_test = functools.partial(_lax_dg, dims)
      expected_output_shape = jax.eval_shape(lax_dg_test, lhs, rhs).shape

      tiling_cfgs = generate_tiling_cfgs(
          lhs.shape, lhs_ca, lhs_ra, rhs.shape, rhs_ca, rhs_ra
      )
      for i in range(len(tiling_cfgs)):
        cfg_in = tiling_cfgs[i]
        tiled_dg_test = functools.partial(_tiled_dg, cfg_in, dims)
        output_shape = jax.eval_shape(tiled_dg_test, lhs, rhs).shape
        msg = (
            'Test failed. Please try the following single test:\n'
            f'lhs_shape = {lhs.shape}\n'
            f'rhs_shape = {rhs.shape}\n'
            'lhs = jnp.ones(lhs_shape)\n'
            'rhs = jnp.ones(rhs_shape)\n'
            f'cfg = {cfg_in}\n'
            f'dims = {dims}\n'
            'out = tiled_dot_general.local_dg(cfg, lhs, rhs, dims)\n'
            'assert out.shape == jax.lax.dot_general(lhs, rhs, dims).shape'
        )
        assert output_shape == expected_output_shape, msg

  def test_tiled_dot_general(self):
    num_ca = 2
    num_ba = 2
    num_lhs_ra = 1
    num_rhs_ra = 2
    max_shape_val = 12
    key = jax.random.PRNGKey(7)
    lhs, rhs, lhs_axes, rhs_axes = generate_inputs(
        key, num_ca, num_ba, num_lhs_ra, num_rhs_ra, max_shape_val
    )
    lhs_ca, lhs_ba, lhs_ra = lhs_axes
    rhs_ca, rhs_ba, rhs_ra = rhs_axes
    dims = ((tuple(lhs_ca), tuple(rhs_ca)), (tuple(lhs_ba), tuple(rhs_ba)))
    expected_output = jax.lax.dot_general(lhs, rhs, dims)

    tiling_cfgs = generate_tiling_cfgs(
        lhs.shape, lhs_ca, lhs_ra, rhs.shape, rhs_ca, rhs_ra
    )
    for i in range(len(tiling_cfgs)):
      cfg_in = tiling_cfgs[i]
      output = tiled_dot_general.tiled_dot_general(cfg_in, lhs, rhs, dims)
      msg = (
          'Test failed. Please try the following single test:\n'
          f'lhs_shape = {lhs.shape}\n'
          f'rhs_shape = {rhs.shape}\n'
          'lhs = jnp.ones(lhs_shape)\n'
          'rhs = jnp.ones(rhs_shape)\n'
          f'cfg = {cfg_in}\n'
          f'dims = {dims}\n'
          'out = tiled_dot_general.local_dg(cfg, lhs, rhs, dims)\n'
          'assert out.shape == jax.lax.dot_general(lhs, rhs, dims).shape'
      )
      # assert output.shape == expected_output.shape, msg
      assert (output == expected_output).all(), msg

  def test_single(self):
    lhs_shape = (31, 22, 9, 8, 27)
    rhs_shape = (8, 9, 22, 27, 7, 12)
    lhs = jnp.ones(lhs_shape)
    rhs = jnp.ones(rhs_shape)
    cfg = Cfg(
        lhs=TensorTiling(
            contraction_axes=[
                AxisTiling(axis=2, tile_count=None, tile_size=9),
                AxisTiling(axis=1, tile_count=None, tile_size=11),
            ],
            remaining_axes=[
                # AxisTiling(axis=0, tile_count=1, tile_size=31)
            ],
        ),
        rhs=TensorTiling(
            contraction_axes=[
                AxisTiling(axis=1, tile_count=None, tile_size=9),
                AxisTiling(axis=2, tile_count=None, tile_size=11),
            ],
            remaining_axes=[
                AxisTiling(axis=5, tile_count=None, tile_size=4),
                # AxisTiling(axis=4, tile_count=7, tile_size=1),
            ],
        ),
    )
    dims = (((2, 1), (1, 2)), ((4, 3), (3, 0)))
    out = tiled_dot_general.tiled_dot_general(cfg, lhs, rhs, dims)
    assert out.shape == jax.lax.dot_general(lhs, rhs, dims).shape

  def test_tiling_state_for_single_tensor(self):
    t_shape = (31, 22, 27, 8, 27)
    t = jnp.ones(t_shape)
    tiled_axes = [
        AxisTiling(axis=1, tile_count=None, tile_size=11),
        AxisTiling(axis=2, tile_count=None, tile_size=9),
    ]
    t_tiling_state = tiled_dot_general.generate_tiling_state(
        t, tiled_axes,
    )
    self.assertEqual(t_tiling_state.untiled_shape, t_shape)
    self.assertEqual(t_tiling_state.tiled_shape, [31, 2, 11, 3, 9, 8, 27])
    self.assertDictEqual(
        t_tiling_state.tile_map, {0: [0], 1: [1, 2], 2: [3, 4], 3: [5], 4: [6]}
    )
    tiled_axes, ca_axes = t_tiling_state.to_tiled_axes_transposed(
        [ax.axis for ax in tiled_axes]
    )
    self.assertEqual(tiled_axes, [1, 3])
    self.assertEqual(ca_axes, [2, 4])

  def test_negative_axis_index(self):
    t_shape = (31, 22, 27, 8, 27)
    t = jnp.ones(t_shape)

    # AxisTiling suppoorts negative axis index
    tiled_axes = [
        AxisTiling(axis=-1, tile_count=None, tile_size=9),
        AxisTiling(axis=-2, tile_count=2, tile_size=None),
    ]
    t_tiling_state = tiled_dot_general.generate_tiling_state(
        t, tiled_axes
    )

    self.assertEqual(t_tiling_state.untiled_shape, t_shape)
    self.assertEqual(t_tiling_state.tiled_shape, [31, 22, 27, 2, 4, 3, 9])
    self.assertEqual(
        t_tiling_state.tile_map, {0: [0], 1: [1], 2: [2], 3: [3, 4], 4: [5, 6]}
    )

    # to_tiled_axes_transposed supports negative axis index
    self.assertEqual(t_tiling_state.to_tiled_axes_transposed([-1]), ([5], [6]))
    self.assertEqual(t_tiling_state.to_tiled_axes_transposed([-2]), ([3], [4]))
    self.assertEqual(t_tiling_state.to_tiled_axes_transposed([-3]), ([], [2]))


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to load a TFRecordDataset and extract TFEvents."""

import collections
import dataclasses
import os
from typing import Dict, List, Optional
import numpy as onp
import tensorflow as tf


@dataclasses.dataclass
class EventSeries:
  """To store series of events as coupled lists. All lists have same length."""
  # Name of the metric
  name: str
  # List of steps when metric was recorded, 1D list of integers.
  steps: onp.ndarray
  # Values recorded at steps, same length as steps, 1D list of floats.
  values: onp.ndarray
  # Time when value was recorded, same length as steps, 1D list of floats.
  wall_times: Optional[onp.ndarray] = None


def _sort_and_deduplicate_entries(event_series: EventSeries) -> EventSeries:
  """Sorts entries by (step, wall_time), deduplicates entries with same step.

  Will keep the one with the latest wall_time and remove wall_time from result.
  If resulting wall_times are not sorted, will raise error.

  Args:
    event_series: Series of (step, value, wall_time) tuples, stored as coupled
      lists in EventSeries dataclass instance. Doesn't need to be sorted, as
      this function will sort it.

  Returns:
    Sorted Series of (step, value) tuples with unique steps, stored as coupled
    lists in EventSeries dataclass instance.
  """
  # Sort lists first by step, then by wall time
  sort_indices = onp.lexsort((event_series.wall_times, event_series.steps))
  event_series.steps = event_series.steps[sort_indices]
  event_series.wall_times = event_series.wall_times[sort_indices]
  event_series.values = event_series.values[sort_indices]
  cur_step = None
  cur_value = None
  res_event_series = EventSeries(
      name=event_series.name,
      steps=onp.array([], dtype=int),
      values=onp.array([]),
      wall_times=onp.array([]))
  for i in range(len(event_series.steps)):
    if event_series.steps[i] != cur_step:
      if cur_value is not None:
        res_event_series.steps = onp.append(res_event_series.steps, cur_step)
        res_event_series.values = onp.append(res_event_series.values, cur_value)
        res_event_series.wall_times = onp.append(res_event_series.wall_times,
                                                 cur_wall_time)
    cur_step = event_series.steps[i]
    cur_value = event_series.values[i]
    cur_wall_time = event_series.wall_times[i]
  if cur_value is not None:
    res_event_series.steps = onp.append(res_event_series.steps, cur_step)
    res_event_series.values = onp.append(res_event_series.values, cur_value)
    res_event_series.wall_times = onp.append(res_event_series.wall_times,
                                             cur_wall_time)

  def _is_sorted(arr: onp.ndarray) -> bool:
    return onp.all(onp.diff(arr) >= 0)

  if not _is_sorted(res_event_series.wall_times):
    raise ValueError(
        'Resulting EventSeries list after sorting by (step, wall_time)'
        ' has unsorted wall_times, likely caused by error in '
        'training.')
  # Remove wall_times data since not needed anymore.
  res_event_series.wall_times = None
  return res_event_series


def get_tfevent_paths(dir_path: str) -> List[str]:
  """Read a TFRecordDataset from a directory path.

  The TFRecordDataset can be loaded from one or multiple files. E.g. if training
  restarts, multiple files would be saved.

  The events file(s) should be located either directly in dir_path, or nested
  one more level (in a subdirectory).

  Args:
    dir_path: Path to a directory containing file(s) readable as a
      TFRecordDataset.

  Returns:
    A list of tf event file paths.
  """
  if not tf.io.gfile.isdir(dir_path):
    raise ValueError(f'dir_path {dir_path} must point to a directory.')

  def _is_events_file(filename: str):
    return 'events.out.tfevents' in filename

  tfevent_file_paths = []
  for filename in tf.io.gfile.listdir(dir_path):
    if tf.io.gfile.isdir(filename):
      for nested_filename in filename.iterdir():
        if _is_events_file(nested_filename):
          tfevent_file_paths.append(
              os.path.join(dir_path, filename, nested_filename))
      break
    else:
      if _is_events_file(filename):
        tfevent_file_paths.append(os.path.join(dir_path, filename))
  return tfevent_file_paths


def get_parsed_tfevents(dir_path: str,
                        tags_to_include: List[str]) -> Dict[str, EventSeries]:
  """Retrieves (step, value) tuples from TFEvents files, filtered by tags.

  Args:
    dir_path: Path to a directory containing TFEvent files.
    tags_to_include: List of tags (attributes), e.g. 'loss' that should be
      included in the report.

  Returns:
    A dictionary mapping tag name to a EventSeries dataclass.
  """
  tfevent_file_paths = get_tfevent_paths(dir_path)

  event_series_dict = collections.defaultdict(list)
  for path in tfevent_file_paths:
    for event in tf.compat.v1.train.summary_iterator(str(path)):
      # There is no v2 API for summary_iterator, so using tf.compat.v1 instead.
      for value in event.summary.value:
        if value.tag in tags_to_include:
          # Some tag names have spaces.
          tag_str = str(value.tag).replace(' ', '_')
          if tag_str not in event_series_dict:
            event_series_dict[tag_str] = EventSeries(
                name=tag_str,
                steps=onp.array([], dtype=int),
                values=onp.array([]),
                wall_times=onp.array([]))
          event_series_dict[tag_str].steps = onp.append(
              event_series_dict[tag_str].steps, event.step)
          event_series_dict[tag_str].values = onp.append(
              event_series_dict[tag_str].values,
              float(tf.make_ndarray(value.tensor)))
          event_series_dict[tag_str].wall_times = onp.append(
              event_series_dict[tag_str].wall_times, event.wall_time)

  for tag_str in event_series_dict:
    # Sort and deduplicate entries with the same step, keeping one with latest
    # wall_time
    event_series_dict[tag_str] = _sort_and_deduplicate_entries(
        event_series_dict[tag_str])
  return event_series_dict



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""utils contains all kinds of small pieces of code used throughout AQT.

Code in this file can't depend on any other AQT file.
However, it is acceptable to grow pieces of funcionality in this file and later
promote them to dedicated files.
"""

import dataclasses
import difflib
import enum
import functools
import pprint
import re
from typing import Any, Sequence, TypeAlias
import flax.struct
import jax
from jax import numpy as jnp


# Some common types.
AxisIdx = int
AxisSize = int

# None means that the template matches any axis size
ShapeTemplate = Sequence[None |int]

# TODO(lew): We should get a better type for jax.lax.dot_general and variants.
DotGeneralT = Any

# Specifies the scale values to search for. Used with `SnrBasedAutoCalibration`
# for auto scale search.
AutoScaleSearchConfig: TypeAlias = Sequence[float]


def assert_shape(shape: Sequence[int], shape_template: ShapeTemplate, msg: str):
  shape = list(shape)
  shape_template = list(shape_template)
  assert len(shape) == len(shape_template), msg
  for axis_size, expected_size in zip(shape, shape_template):
    assert (
        expected_size is None or axis_size == expected_size
    ), f'{msg}: {shape} does not match {shape_template}'


def assert_eq(value: Any, expected: Any, value_name: str):
  msg = f'{value_name}={value}, {expected=}'
  assert value == expected, msg


flax_slots_dataclass = functools.partial(
    flax.struct.dataclass, frozen=False, slots=True
)

# TODO: b/332418233 - Consolidate flax_slots_kw_only_dataclass and
# flax_slots_dataclass into a single decorator once the migration is
# done. This will exist only temporarily while completing the
# migration.
flax_slots_kw_only_dataclass = functools.partial(
    flax_slots_dataclass, kw_only=True
)


def dataclass_field(default='no default value'):
  # We can't use None as the default's default,
  # because someone actually might want to have None as default of the field.
  if default == 'no default value':
    return dataclasses.field()
  else:
    return dataclasses.field(default_factory=default)


class QuantMode(enum.Enum):
  TRAIN = 1
  CALIBRATE = 2
  CONVERT = 3
  SERVE = 4


def static_field(**kwargs):
  return flax.struct.field(pytree_node=False, **kwargs)


def dynamic_field(**kwargs):
  return flax.struct.field(pytree_node=True, **kwargs)


def print_diff(str_a: str, str_b: str, do_print_diff=False):
  if do_print_diff:
    diff_generator = difflib.context_diff(str_a.split(' '), str_b.split(' '))
    print('Diff:')
    for diff in diff_generator:
      print(diff)
  print(f'first string (actual):\n{str_a}')


def test_pprint_eq(
    input_a: Any, input_b: Any, remove_memory_addresses: bool = False
):
  str_input_a = input_a if isinstance(input_a, str) else pprint.pformat(input_a)
  str_input_b = input_b if isinstance(input_b, str) else pprint.pformat(input_b)
  if remove_memory_addresses:
    str_input_a = re.sub(r' at 0x.*>', '>', str_input_a, 0, re.MULTILINE)
    str_input_b = re.sub(r' at 0x.*>', '>', str_input_b, 0, re.MULTILINE)
  assert str_input_a == str_input_b, print_diff(str_input_a, str_input_b)


def infer_dtype_from_bits(bits: int) -> None | jnp.dtype:
  """Get the dtype for the number of bits provided.

  Args:
    bits: number of bits for the dtype.

  Returns:
    The corresponding container dtype for the number of bits provided.
  """
  if bits == 4:
    return jnp.int4
  else:
    if bits <= 8 and bits >= 2:
      return jnp.int8
    else:
      return None


def get_remaining_axes(
    rank: int,
    contraction_axes: Sequence[AxisIdx],
    batch_axes: Sequence[AxisIdx],
):
  """Returns the remaining axes."""
  ret = []
  for i in range(rank):
    if i not in list(contraction_axes) + list(batch_axes):
      ret.append(i)
  return ret


@flax_slots_dataclass
class Context:
  key: None | jax.Array = dynamic_field()
  train_step: None | int = dynamic_field()
  quant_mode: QuantMode = static_field(default=QuantMode.TRAIN)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for conv_general."""

import itertools
from typing import Iterable, Optional

from absl.testing import absltest
from absl.testing import parameterized
from aqt.common import aqt_config
from aqt.jax import aqt_conv_general
from aqt.jax import aqt_ops
from aqt.jax import aqt_tensor
from aqt.test import aqt_conv_test_base
from flax import linen as nn
import jax
from jax import lax
import jax.numpy as jnp


def generate_nondefault_conv_args():
  lhs_specs = ["NHWC", "HNWC", "HWNC", "HWCN"]
  rhs_specs = ["OHWI", "HOWI", "HWOI", "HWIO"]
  out_specs = ["CNHW", "NCHW", "NHCW", "NHWC"]

  case_dicts = []
  for lhs_spec, rhs_spec, out_spec in itertools.product(lhs_specs, rhs_specs,
                                                        out_specs):
    case_dicts.append(
        dict(
            dimension_numbers=(lhs_spec, rhs_spec, out_spec),
            window_strides=[1, 1],
            padding="VALID"))

  window_strides = [(1, 1), (2, 2), (1, 2), (2, 1)]
  paddings = ["SAME", "VALID", ((1, 2), (2, 1)), ((2, 2), (2, 2))]
  for strides, padding in itertools.product(window_strides, paddings):
    case_dicts.append(
        dict(
            dimension_numbers=("NHWC", "HWIO", "NWHC"),
            window_strides=strides,
            padding=padding))

  return case_dicts


class ConvGeneralModule(nn.Module):
  lhs_config: Optional[aqt_config.AqtScheduleConfig]
  rhs_config: Optional[aqt_config.AqtScheduleConfig]
  lhs_shape: Iterable[int]
  rhs_shape: Iterable[int]

  @nn.compact
  def __call__(self,
               lhs,
               rhs,
               weights,
               event_count,
               event_count_for_filter,
               train=True,
               **kwargs):
    lhs_quantizer = aqt_tensor.TensorQuantizer(
        list(self.lhs_shape), self.lhs_config)
    rhs_quantizer = aqt_tensor.TensorQuantizer(
        list(self.rhs_shape), self.rhs_config)

    lhs_quantizer.update(lhs, weights, event_count)
    rhs_quantizer.update(rhs, weights, event_count_for_filter)

    def conv_op(lhs, rhs):
      return aqt_ops.aqt_conv_general_dilated(
          lhs, rhs, lhs_quantizer, rhs_quantizer, train=train, **kwargs)

    return conv_op


class ConvGeneralTest(aqt_conv_test_base.ConvTest):

  def conv_op_quantized(
      self,
      input,  # pylint: disable=redefined-builtin
      filter,  # pylint: disable=redefined-builtin
      input_config,
      filter_config,
      event_count,
      event_count_for_filter=None,
      input_weights=None,
      train=True,
      var_scope_name=None,
      **kwargs):
    event_count_for_filter = event_count_for_filter if event_count_for_filter else event_count

    input = self.constant(input)
    filter = self.constant(filter)

    module = ConvGeneralModule(input_config, filter_config, input.shape,
                               filter.shape)
    conv_general, _ = module.init_with_output(
        jax.random.PRNGKey(0), input, filter, input_weights, event_count,
        event_count_for_filter, train, **kwargs)

    return conv_general(input, filter)

  def conv_op_unquantized(self, input, filter, **kwargs):  # pylint: disable=redefined-builtin
    input = self.constant(input)
    filter = self.constant(filter)
    return lax.conv_general_dilated(input, filter, **kwargs)

  def get_conv_kwargs(
      self,
      strides,
      padding,
      data_format="NHWC",
      dilations=None,
      lhs_dilation=None):
    return dict(
        window_strides=(strides, strides),
        padding=padding,
        lhs_dilation=lhs_dilation,
        rhs_dilation=dilations,
        dimension_numbers=(data_format, "HWIO", data_format),
        feature_group_count=1,
        batch_group_count=1)

  def constant(self, x):
    return jnp.array(x)

  def gradients(self, fwd_func, x, w):
    y, vjp_fn = jax.vjp(fwd_func, x, w)
    return vjp_fn(jnp.ones(y.shape))[0]

  def test_vars_over_inputs_at_inference(self):
    input_config, x, filter_config, w = self.exact_int8_conv_example(
        lhs_use_quantized_variable=True, rhs_use_quantized_variable=True)

    x = self.constant(x)
    w = self.constant(w)
    kwargs = self.get_conv_kwargs(strides=1, padding="VALID")

    module = ConvGeneralModule(input_config, filter_config, x.shape, w.shape)

    # Since quantized variables are not used at training, the conv_general with
    # zero inputs should produce zero values.
    conv_general, state = module.init_with_output(
        jax.random.PRNGKey(0),
        x,
        w,
        weights=None,
        event_count=0,
        event_count_for_filter=0,
        train=True,  # in training
        **kwargs)
    actual = conv_general(jnp.zeros_like(x), jnp.zeros_like(w))
    expected = jnp.zeros_like(actual)
    self.assertAllEqual(actual, expected)

    # Since quantized variables should be always used at inference, the
    # conv_general will rely on quantized variables.
    conv_general_infer, _ = module.apply(
        state,
        x,
        w,
        weights=None,
        event_count=0,
        event_count_for_filter=0,
        train=False,  # in inference
        **kwargs,
        mutable=True)
    actual = conv_general_infer(jnp.zeros_like(x), jnp.zeros_like(w))
    expected = conv_general(x, w)
    self.assertAllEqual(actual, expected)

  def test_zero_preservation_for_input_dilation(self):
    """Validates preserve_zero is set to True when the filter is dilated."""
    input_config = aqt_conv_test_base.schedule_config(
        "input", bits=8, const_coeff=1.0, preserve_zero=False)
    filter_config = aqt_conv_test_base.schedule_config(
        "filter", bits=8, const_coeff=1.0)

    x, w = self.create_random_input_and_filter()

    kwargs = self.get_conv_kwargs(
        strides=1, padding=((1, 1), (1, 1)), lhs_dilation=[2, 2])

    with self.assertRaisesRegex(aqt_config.ConfigError,
                                "must be True if the input is dilated"):
      self.conv_op_quantized(
          x,
          w,
          input_config,
          filter_config,
          event_count=0,
          **kwargs)

  def test_jvp(self):
    input_config, x, filter_config, w = self.exact_int8_conv_example()

    kwargs = self.get_conv_kwargs(strides=1, padding="VALID")

    def actual_fwd(x, w):
      return self.conv_op_quantized(
          x, w, input_config, filter_config, event_count=0, **kwargs)

    def expected_fwd(x, w):
      return self.conv_op_unquantized(x, w, **kwargs)

    actual = lambda x, w: jnp.sum(actual_fwd(x, w)**2)
    expected = lambda x, w: jnp.sum(expected_fwd(x, w)**2)

    _, y_dot_actual = jax.jvp(actual, (x, w),
                              (jnp.ones(x.shape), jnp.ones(w.shape)))
    _, y_dot_expected = jax.jvp(expected, (x, w),
                                (jnp.ones(x.shape), jnp.ones(w.shape)))
    self.assertAllEqual(y_dot_actual, y_dot_expected)

  @parameterized.parameters(generate_nondefault_conv_args())
  def test_nondefault_conv_args(self, dimension_numbers, window_strides,
                                padding):
    def get_share_stats_axes(dimension_numbers):
      # Based on the dimension number of string format, create share_stats_axes
      # which should be provided to stats config. Note that the first dimension
      # is always non-contracting one: "N" for LHS and "O" for RHS.
      return sorted(dimension_numbers[1:])

    # Create an input tensor of the shape (2, 3, 3, 1) and a filter tensor of
    # the shape (2, 2, 1, 2) with ("NHWC", "HWIO, "NHWC") dimension number.
    # In order to effectively see that the inv_scales are applied to conv result
    # correctly no matter which dimension numbers are specified, each of the
    # batch/feature dimensions should have different scales.
    x = jnp.array(
        [[[12.6, -23.1, -4.2], [-40.3, 255.0, 0.0], [0.0, -32.5, 55.1]],
         [[63.75, -47.3, -1.3], [-24.2, 14.4, 0.0], [0.0, -29.4, 30.1]]],
        dtype=jnp.float32)[:, :, :, jnp.newaxis]
    w0 = jnp.array([[1, 7.5], [-5.5, -3]], dtype=jnp.float32)
    w1 = jnp.array([[15, -2.4], [-1.1, 1.2]], dtype=jnp.float32)
    w = jnp.stack([w0, w1])[:, :, jnp.newaxis, :]

    # 8 bits quantization with max_dev_coeff=1
    input_config = aqt_conv_test_base.schedule_config(
        "input", const_coeff=0, bits=8, max_dev_coeff=1)
    # 2 bits quantization with max_dev_coeff=1
    filter_config = aqt_conv_test_base.schedule_config(
        "filter", const_coeff=0, bits=2, max_dev_coeff=1)

    dimension_numbers = lax.conv_dimension_numbers(x.shape, w.shape,
                                                   dimension_numbers)
    # Transpose the input and filter according to given dimension numbers.
    dim_numbers = {
        "NHWC": (0, 3, 1, 2),
        "HWIO": (3, 2, 0, 1),
    }
    x = aqt_conv_general._transpose_inv_scale(x, dim_numbers["NHWC"],
                                              dimension_numbers[0])
    w = aqt_conv_general._transpose_inv_scale(w, dim_numbers["HWIO"],
                                              dimension_numbers[1])
    input_config.stats_config.share_stats_axes = get_share_stats_axes(
        dimension_numbers[0])
    filter_config.stats_config.share_stats_axes = get_share_stats_axes(
        dimension_numbers[1])

    kwargs = dict(
        window_strides=window_strides,
        padding=padding,
        dimension_numbers=dimension_numbers)

    actual = self.conv_op_quantized(
        x, w, input_config, filter_config, event_count=0, **kwargs)

    # The quantized tensors for the examples created above.
    x = jnp.array([[[6, -12, -2], [-20, 127, 0], [0, -16, 28]],
                   [[127, -95, -3], [-48, 29, 0], [0, -59, 60]]],
                  dtype=jnp.float32)[:, :, :, jnp.newaxis]
    w0 = jnp.array([[0, 1], [-1, -1]], dtype=jnp.float32)
    w1 = jnp.array([[1, 0], [0, 0]], dtype=jnp.float32)
    w = jnp.stack([w0, w1])[:, :, jnp.newaxis, :]

    x = aqt_conv_general._transpose_inv_scale(x, dim_numbers["NHWC"],
                                              dimension_numbers[0])
    w = aqt_conv_general._transpose_inv_scale(w, dim_numbers["HWIO"],
                                              dimension_numbers[1])

    expected = self.conv_op_unquantized(x, w, **kwargs)

    batch_inv_scale = jnp.array([2, 0.5])[:, jnp.newaxis, jnp.newaxis,
                                          jnp.newaxis]
    batch_inv_scale = aqt_conv_general._transpose_inv_scale(
        batch_inv_scale, dim_numbers["NHWC"], dimension_numbers[2])
    feature_inv_scale = jnp.array([10, 5])[jnp.newaxis, jnp.newaxis,
                                           jnp.newaxis, :]
    feature_inv_scale = aqt_conv_general._transpose_inv_scale(
        feature_inv_scale, dim_numbers["NHWC"], dimension_numbers[2])
    expected *= (batch_inv_scale * feature_inv_scale)

    self.assertAllEqual(actual, expected)

  def test_weight_only_quantization(self):
    x, w = self.create_random_input_and_filter()
    input_config = aqt_conv_test_base.schedule_config(
        "input", bits=8, const_coeff=1.0)
    filter_config = aqt_conv_test_base.schedule_config(
        "filter", bits=8, const_coeff=1.0)
    input_config.tensor_configs[0].quant_config = aqt_config.FloatConfig()

    kwargs = self.get_conv_kwargs(strides=1, padding="VALID")
    result_with_float_config = self.conv_op_quantized(
        x, w, input_config, filter_config, event_count=0, **kwargs)
    result_with_none_config = self.conv_op_quantized(
        x, w, None, filter_config, event_count=0, **kwargs)

    self.assertAllEqual(result_with_float_config, result_with_none_config)


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Small-sized unquantized model.

A 'small size' halves each model size parameter of the full model:
* 3 layers in the encoder and decoder
* 512 emd_dim
* 8 heads
* 512 qkv_dim
* 2048 mlp_dim
"""

from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs import base_config


def get_config(quant_target=base_config.QuantTarget.NONE):
  """Returns configuration for a small transformer model."""
  config = base_config.get_config(n_layers=3, quant_target=quant_target)
  model = config.model_hparams
  model.emb_dim = 512
  model.num_heads = 8
  model.qkv_dim = 512
  model.mlp_dim = 2048
  config.metadata.hyper_str = 'small_bfloat16'
  return config



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines a reasonable baseline configuration for Resnet.

Specific experimental configurations should override these values as needed.
"""

import enum
from typing import List

from aqt.jax_legacy.jax.imagenet.configs_script import config_schema


class ImagenetType(enum.Enum):
  """Enum to distinguish between different resnet architectures.

  Details of the architecture is at ../imagenet.png from
  He *et al.*, 2015 https://arxiv.org/pdf/1512.03385.pdf.
  """
  RESNET29 = enum.auto()
  RESNET50 = enum.auto()
  RESNET101 = enum.auto()
  RESNET152 = enum.auto()
  RESNET200 = enum.auto()

  def get_residual_layers(self) -> List[int]:
    """Returns number of resudual blocks corresponding to the architecture."""
    if self == self.RESNET29:
      return [2, 2, 3, 2]
    elif self == self.RESNET50:
      return [3, 4, 6, 3]
    elif self == self.RESNET101:
      return [3, 4, 23, 3]
    elif self == self.RESNET152:
      return [3, 8, 36, 3]
    elif self == self.RESNET200:
      return [3, 24, 36, 3]
    else:
      raise ValueError(f"ImagenetType {self.value} is unknown.")


# TODO(shivaniagrawal): share the QuantTargets
class QuantTarget(enum.Enum):
  """Specifies which part of the model to quantize."""
  # No quantization
  NONE = enum.auto()

  # Quantize only the model weights. Activations are left unquantized.
  WEIGHTS_ONLY = enum.auto()

  # Quantize all the weights and activations. Activations use fixed
  # pre-specified clipping bounds.
  WEIGHTS_AND_FIXED_ACT = enum.auto()

  # Quantize all the weights and activations. Activations use clipping bounds
  # derived from activation statistics collected during training.
  WEIGHTS_AND_AUTO_ACT = enum.auto()


def get_base_config(imagenet_type: ImagenetType, quant_target: QuantTarget):
  """Returns config that sets hyperparameters common to all quant targets.

  Fields in that config can then be overridden to customize a configuration.

  Note that two hyperparameters, model architecture kind and whether to
  automatically find clipping bounds for activations, have to be specified in
  advance as keyword arguments to this function instead of being overridden in
  the returned configdict. That is because these parameters affect the name and
  number of fields in the configdict instance, which can't be changed after
  creation: there will be one set of overridable parameters per layer in the
  configdict, and the field names in the 'quant_act' fields change depending on
  'quant_target'.

  Args:
    imagenet_type: Resnet model architecture.
    quant_target: Given quantization target, helpful for making decision whether
      to get config for automatic bounds alculation for activations or fixed
      bounds.

  Returns:
    A ConfigDict instance suitable for WMT training.
  """
  resnet_layers = imagenet_type.get_residual_layers()
  num_blocks = sum(resnet_layers)

  use_auto_acts = True if quant_target == QuantTarget.WEIGHTS_AND_AUTO_ACT else False
  config = config_schema.get_config(
      num_blocks=num_blocks, use_auto_acts=use_auto_acts)
  config.update({
      "base_learning_rate": 0.1,
      "momentum": 0.9,
      "weight_decay": 0.0001,
      "activation_bound_update_freq": -1,
      "activation_bound_start_step": -1,
      "prec": None,

      "sparsity": {
          "type": "STRUCTURED_NM",
          "prune_rate": None,
          "order": "C",
          "absolute": True,
          "smallest": True,
          "structure_decay": False,
          "sparse_ste": False
      },
      "quant_type": "fake_quant",
      "weight_quant_granularity": "per_channel",
      "act_function": "relu",
      "shortcut_ch_shrink_method": "none",
      "shortcut_ch_expand_method": "none",
      "shortcut_spatial_method": "none",
      "lr_scheduler": {
          "warmup_epochs": 5,
          "cooldown_epochs": 50,
          "scheduler": "cosine",
          "num_epochs": 250,
          "endlr": 0.0,
          "knee_lr": 1e-5,
          "knee_epochs": 125,
      },
      "optimizer": "sgd",
      "adam": {
          "beta1": 0.9,
          "beta2": 0.999
      },
      "early_stop_steps": -1,  # -1 means no early stop
      "weight_quant_start_step": 0,  # 0 means turned on by default
      "teacher_model": "labels",
      "is_teacher": True,  # by default train the vanilla resnet
      "seed": 0,
  })

  proj_layers = [sum(resnet_layers[:x]) for x in range(len(resnet_layers))]
  for idx in range(num_blocks):
    if idx not in proj_layers:
      config.model_hparams.residual_blocks[idx].conv_proj = None
      config.model_hparams.residual_blocks[
          idx].conv_1.quant_act.input_distribution = "positive"

  config.model_hparams.filter_multiplier = 1.
  config.model_hparams.se_ratio = 0.5
  config.model_hparams.init_group = 32
  config.half_shift = False

  return config


def get_auto_acts_config(imagenet_type: ImagenetType):
  """Returns config appropriate for automatic activation quantization."""
  config = get_base_config(
      imagenet_type=imagenet_type,
      quant_target=QuantTarget.WEIGHTS_AND_AUTO_ACT)
  config.quant_act.bounds.update({  # update default values for auto acts
      "initial_bound": -1.0,
      "stddev_coeff": 3.0,
      "absdev_coeff": 3.0,
      "mix_coeff": 0.0,
      "reset_stats": False,
      "granularity": "per_channel",
      "ema_coeff": 0.1,
      "use_cams": False,
      "exclude_zeros": True,
      "use_mean_of_max": True,
      "use_old_code": True,
  })
  config.activation_bound_start_step = 7500
  config.activation_bound_update_freq = -1
  return config


def get_weights_only_config(imagenet_type: ImagenetType):
  """Returns config for weights-only quantization."""
  config = get_base_config(
      imagenet_type=imagenet_type, quant_target=QuantTarget.WEIGHTS_ONLY)
  config.quant_act.bounds = None
  return config


def get_fixed_acts_config(imagenet_type: ImagenetType):
  """Returns config for activation quantization with fixed bounds."""
  config = get_base_config(
      imagenet_type=imagenet_type,
      quant_target=QuantTarget.WEIGHTS_AND_FIXED_ACT)
  config.quant_act.bounds = 6.0
  return config


def get_config(quant_target: QuantTarget, imagenet_type: ImagenetType):
  """Returns config for a given quantization target and layer count."""
  if quant_target == QuantTarget.WEIGHTS_ONLY:
    return get_weights_only_config(imagenet_type=imagenet_type)
  elif quant_target == QuantTarget.WEIGHTS_AND_FIXED_ACT:
    return get_fixed_acts_config(imagenet_type=imagenet_type)
  elif quant_target == QuantTarget.WEIGHTS_AND_AUTO_ACT:
    return get_auto_acts_config(imagenet_type=imagenet_type)
  elif quant_target == QuantTarget.NONE:
    return get_base_config(
        imagenet_type=imagenet_type, quant_target=quant_target)
  else:
    raise ValueError(f"quant_target {quant_target} not understood.")



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Numerics for fp8."""

from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import numerics
import jax
import jax.numpy as jnp


float_repr = jax.Array


@utils.flax_slots_kw_only_dataclass
class FpNumericsConfig:
  nexp: int
  minexp: int
  nmant: int
  has_subnormals: bool
  has_two_nan: bool
  has_naninf: bool
  radix: int

##########
## FP6
##########
e5m0_ocp = FpNumericsConfig(
    nexp=5,
    minexp=0,
    nmant=0,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

e4m1_ocp = FpNumericsConfig(
    nexp=4,
    minexp=0,
    nmant=1,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

e3m2_ocp = FpNumericsConfig(
    nexp=3,
    minexp=0,
    nmant=2,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

e2m3_ocp = FpNumericsConfig(
    nexp=2,
    minexp=0,
    nmant=3,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

e1m4_ocp = FpNumericsConfig(
    nexp=1,
    minexp=0,
    nmant=4,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

e0m5_ocp = FpNumericsConfig(
    nexp=0,
    minexp=0,
    nmant=5,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

##########
## FP5
##########
e4m0_ocp = FpNumericsConfig(
    nexp=4,
    minexp=0,
    nmant=0,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

e3m1_ocp = FpNumericsConfig(
    nexp=3,
    minexp=0,
    nmant=1,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

e2m2_ocp = FpNumericsConfig(
    nexp=2,
    minexp=0,
    nmant=2,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

e1m3_ocp = FpNumericsConfig(
    nexp=1,
    minexp=0,
    nmant=3,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

e0m4_ocp = FpNumericsConfig(
    nexp=0,
    minexp=0,
    nmant=4,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

##########
## FP4
##########
e3m0 = FpNumericsConfig(
    nexp=3,
    minexp=0,
    nmant=0,
    has_subnormals=False,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

e3m0_ocp = FpNumericsConfig(
    nexp=3,
    minexp=0,
    nmant=0,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

e2m1 = FpNumericsConfig(
    nexp=2,
    minexp=0,
    nmant=1,
    has_subnormals=False,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

e2m1_ocp = FpNumericsConfig(
    nexp=2,
    minexp=0,
    nmant=1,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

e1m2 = FpNumericsConfig(
    nexp=1,
    minexp=0,
    nmant=2,
    has_subnormals=False,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

e1m2_ocp = FpNumericsConfig(
    nexp=1,
    minexp=0,
    nmant=2,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

e0m3 = FpNumericsConfig(
    nexp=0,
    minexp=0,
    nmant=3,
    has_subnormals=False,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

e0m3_ocp = FpNumericsConfig(
    nexp=0,
    minexp=0,
    nmant=3,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

##########
## Misc
##########
e1m0 = FpNumericsConfig(
    nexp=1,
    minexp=0,
    nmant=0,
    has_subnormals=False,
    has_two_nan=False,
    has_naninf=False,
    radix=2,
)

float8_e4m3fn = FpNumericsConfig(
    nexp=4,
    minexp=-6,
    nmant=3,
    has_subnormals=True,
    has_two_nan=True,
    has_naninf=False,
    radix=2,
)

float8_e5m2 = FpNumericsConfig(
    nexp=5,
    minexp=-14,
    nmant=2,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=True,
    radix=2,
)

float16 = FpNumericsConfig(
    nexp=5,
    minexp=-14,
    nmant=10,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=True,
    radix=2,
)

RADIX4 = FpNumericsConfig(
    nexp=3,
    minexp=0,
    nmant=0,
    has_subnormals=True,
    has_two_nan=False,
    has_naninf=False,
    radix=4,
)


def fp_round(
    x,
    *,
    cfg: FpNumericsConfig,
    key: jax.Array,
    stochastic_rounding: bool,
    test_noise_axis=None,
):
  """Round function dispatcher."""
  assert cfg.radix == 2 or cfg.radix == 4, 'Only radix 2 and 4 are supported.'
  round_fn = radix2_round if cfg.radix == 2 else radix4_round
  return round_fn(
      x,
      cfg=cfg,
      key=key,
      stochastic_rounding=stochastic_rounding,
      test_noise_axis=test_noise_axis,
  )


# TODO(lew): Add minexp.
def radix2_round(
    x,
    *,
    cfg: FpNumericsConfig,
    key: jax.Array,
    stochastic_rounding: bool,
    test_noise_axis=None,
):
  """FP stochastic rounding for a given mantissa and exponent. Returns bf16."""
  nexp = cfg.nexp
  nmant = cfg.nmant
  assert cfg.minexp == 0, 'minexp not implemented'
  assert not cfg.has_two_nan, 'two_nan not implemented'
  assert not cfg.has_naninf, 'naninf not implemented'

  orig_x = x
  input_dtype = x.dtype
  total_bits = jnp.finfo(input_dtype).bits
  # (sign, exp, mantissa) = (1,8,x)
  msg = f'Unsupported dtype for sthochastic rounding: {input_dtype}.'
  assert input_dtype == jnp.bfloat16 or input_dtype == jnp.float32, msg
  # bf16 bit pattern: seeeeeeeemmmmmmm
  #                   7654321076543210
  container_man = jnp.finfo(input_dtype).nmant
  bits_dtype = jnp.uint16 if total_bits == 16 else jnp.uint32
  assert nmant <= container_man, 'bf16 man bits'
  man_trunc_bits = container_man - nmant  # e2m1 in bf16: 7-1
  # example e2m1 in bf16: man_trunc_mask = 0b0000000000111111  (7-1 1s)
  man_trunc_mask = bits_dtype((1 << man_trunc_bits) - 1)  # 0b00000000_00111111

  # TODO(lew): We can use smaller dtype
  x = jax.lax.bitcast_convert_type(x, bits_dtype)  # adds noise to man
  if stochastic_rounding:
    if test_noise_axis is not None:
      noise_axis_size = x.shape[test_noise_axis]
      sh = list((1,) * len(x.shape))
      sh[test_noise_axis] = noise_axis_size
      rnd_bits = jnp.arange(noise_axis_size, dtype=bits_dtype).reshape(sh)
    else:
      rnd_bits = jax.random.bits(key, x.shape, bits_dtype)
    noise = jax.lax.convert_element_type(rnd_bits, bits_dtype) & man_trunc_mask
  else:
    # example e2m1 in bf16: noise = 0b0000000000100000  (shift = 7-1 - 1)
    noise = 1 << (man_trunc_bits - 1)  #  represents 0.5 in the container
    noise = jax.lax.convert_element_type(noise, bits_dtype)
  # This noise add might overflow up to sign bit if x(bf16) has max exp.
  # In bf16 this happens if x=nan or x=inf.
  x = x + noise  # uint addition, potential carry into exponent, this is correct
  x = jax.lax.bitwise_and(x, ~man_trunc_mask)  # zero-out lowest "6" bits in man

  if nexp is not None:
    # We use 1 as a minimal representable value. TODO(lew): use cfg.minexp.
    # We will clip to [1, 2^(2**nexp - 1) * (2-1 ^ nmant)] interval.
    assert input_dtype == jnp.bfloat16
    # min_exp = jnp.uint16(0b0_10000000_0000000)
    # max_exp = jnp.uint16(0b0_10000111_0000000)
    # bf16_exp_bias = jnp.uint16(0b0_01111111)
    min_exp = jnp.uint16(0b0_01111111) - (1 if cfg.has_subnormals else 0)
    max_exp = min_exp + 2**nexp - 1

    min_repr = min_exp << container_man
    # min_repr_float is used to select which SR to use.
    min_repr_float = jax.lax.bitcast_convert_type(min_repr, input_dtype)
    # container_mant_mask is 0b0000_0000_0111_1111
    container_mant_mask = bits_dtype((1 << container_man) - 1)
    # max_mant for e2m1 in bf16: 0b0000_0000_0100_0000
    max_mant = jnp.uint16(container_mant_mask & ~man_trunc_mask)
    max_repr = (max_exp << container_man) + max_mant

    sign_mask = jnp.uint16(0b1_00000000_0000000)
    sign = jax.lax.bitwise_and(x, sign_mask)
    abs_x = jax.lax.bitwise_and(x, ~sign_mask)
    # TODO(lew): we can do faster clipping with bit twiddling
    # Clipping of positive values in int domain is the same as in float.
    clip_x = jnp.clip(abs_x, min_repr, max_repr)  # sign bit is 0
    x = jax.lax.bitwise_or(sign, clip_x)
    x = jax.lax.bitcast_convert_type(x, input_dtype)
    if cfg.has_subnormals:
      # In simulated numerics, what We would like to represent as a subnormal is
      #    sign * 2^min_exp * (0.mantissa_s) = sign * (0.mantissa_s)
      # However, in container format, the same number is represented by
      #    sign * 2^(min_exp-1) * (1.mantissa) = sign * 0.5 * (1.mantissa_c)
      # Note that they are different in bit repr and mantissa_s != mantissa_c.
      # So we cannot directly truncate mantissa_c to get mantissa_s.
      # Also note that the container subnormal is never hit when simulating
      # the low-precision numerics.
      # Alternatively, we can simply do integer rounding inside the subnormal
      # regime since it is linear.
      orig_x_bit_repr = jax.lax.bitcast_convert_type(orig_x, bits_dtype)
      orig_abs_x = jax.lax.bitwise_and(orig_x_bit_repr, ~sign_mask)
      orig_clip_x = jnp.clip(orig_abs_x, min_repr, max_repr)
      exp = jax.lax.bitwise_and(orig_clip_x, ~container_mant_mask)
      if stochastic_rounding:
        if test_noise_axis is not None:
          noise_axis_size = x.shape[test_noise_axis]
          sh = list((1,) * len(x.shape))
          sh[test_noise_axis] = noise_axis_size
          rnd_bits = jnp.arange(noise_axis_size, dtype=bits_dtype).reshape(sh)
        else:
          rnd_bits = jax.random.bits(key, x.shape, bits_dtype)
        rnd_bits = (jnp.uint32(rnd_bits) << 7) | jnp.float32(1).view(jnp.uint32)
        noise = (
            jax.lax.bitcast_convert_type(rnd_bits, jnp.float32)
            - 1.5
            + 2 ** (-17)  # to make noise symmetric to zero
        )
      else:
        noise = 0.0
      # SR within subnormal is adding random noise [-0.5, 0.5) then round.
      subnormal = (jnp.round(orig_x * 2**nmant + noise) / 2**nmant).astype(
          input_dtype
      )
      x = jnp.where(exp == (min_exp << container_man), subnormal, x)
    else:
      # Rounding of values in [-1, 1] interval follows different logic.
      if stochastic_rounding:
        assert bits_dtype == jnp.uint16
        if test_noise_axis is not None:
          noise_axis_size = x.shape[test_noise_axis]
          sh = list((1,) * len(x.shape))
          sh[test_noise_axis] = noise_axis_size
          rnd_bits = jnp.arange(noise_axis_size, dtype=bits_dtype).reshape(sh)
          # Use 16-bit noise for deterministic SR tests.
          # Note: adding left shift operation can adjust the effective noise bit
          # For example rnd_bits = rnd_bits << 8 means using noise with
          # (bits_dtype - 8) effective bits.
          # The effective noise bit should be consistent with the number of
          # inputs in determistic SR test, e.g., 8-bit noise requires 2**8
          # inputs to the test.
        else:
          # Note: bits_dtype affects the effective noise bit in the
          # "test_noise_axis is not None" branch. Remember to change the shift
          # operation and the number of inputs to the deterministic SR test.
          rnd_bits = jax.random.bits(key, x.shape, bits_dtype)
        rnd_bits = (jnp.uint32(rnd_bits) << 7) | jnp.float32(2).view(jnp.uint32)
        # rnd_bits here is 32 bits, with exponent 10000000, first 8 mantissa
        # being the generated "noise" and the rest of mantissa being 0.
        # Sign bit is 0.
        rnd_floats = jax.lax.bitcast_convert_type(rnd_bits, jnp.float32) - (
            3.0 - 2 ** (-16)
        )
        # rx is 1 or -1 - SR of orig_x
        rx = (orig_x > rnd_floats) * 2 - 1.0
        x = jnp.where(jnp.abs(orig_x) < min_repr_float, rx, x)
    # Deal with overflow
    max_subnormal = jax.lax.convert_element_type(
        2**cfg.minexp * (1 - 2 ** (-nmant)), input_dtype
    )
    if nexp == 0 and cfg.has_subnormals:
      # Clip to max subnormal when there are no exponent bits and subnormal
      # presents in the numerics.
      x = jnp.clip(x, -max_subnormal, max_subnormal)
  else:
    x = jax.lax.bitcast_convert_type(x, input_dtype)

  return x


def fp_round_new(
    x,
    *,
    cfg: FpNumericsConfig,
    key: jax.Array,
    stochastic_rounding: bool,
    test_noise_axis=None,
):
  """FP stochastic rounding for a given mantissa and exponent. Returns bf16."""
  # This function is identical to fp_round but with improved readability.
  # Information about the numerics to be simulated
  nexp = cfg.nexp
  nmant = cfg.nmant
  assert cfg.minexp == 0, 'minexp not implemented'
  assert not cfg.has_two_nan, 'two_nan not implemented'
  assert not cfg.has_naninf, 'naninf not implemented'

  # Information about the container
  # BF16 bit pattern: (sign, exp, mantissa) = (1,8,x)
  # Notation: ctner = container
  input_dtype = x.dtype
  msg = f'Unsupported dtype for sthochastic rounding: {input_dtype}.'
  assert input_dtype == jnp.bfloat16, msg
  ctner_total_bits = jnp.finfo(input_dtype).bits
  ctner_nmant = jnp.finfo(input_dtype).nmant
  assert nmant <= ctner_nmant, 'bf16 man bits'
  bits_dtype = jnp.uint16

  # Masks
  ctner_mant_mask = bits_dtype((1 << ctner_nmant) - 1)  # 0_00000000_1111111
  mant_trunc_bits = ctner_nmant - nmant
  # Mantissa truncation mask:
  #     exp    nmant
  # 0_00000000_0..0011
  mant_trunc_mask = bits_dtype((1 << mant_trunc_bits) - 1)
  sign_mask = bits_dtype(0b1 << (ctner_total_bits - 1))

  min_normal = jax.lax.convert_element_type(2**cfg.minexp, input_dtype)
  min_normal_repr = jax.lax.bitcast_convert_type(min_normal, bits_dtype)
  bias = 1 if cfg.has_subnormals else 0
  ctner_bias = 127
  # TODO(yichizh): min_ctner_exp should be (ctner_bias - bias - cfg.min_exp)
  min_ctner_exp = bits_dtype(ctner_bias - bias)
  max_ctner_exp = bits_dtype(min_ctner_exp + 2**nexp - 1)
  max_ctner_mant = bits_dtype(ctner_mant_mask & ~mant_trunc_mask)
  max_normal_repr = (max_ctner_exp << ctner_nmant) + max_ctner_mant
  max_normal = jax.lax.bitcast_convert_type(max_normal_repr, input_dtype)
  max_subnormal = jax.lax.convert_element_type(
      2**cfg.minexp * (1 - 2 ** (-nmant)), input_dtype
  )

  def trunc_and_clip_normal(x_in: float_repr) -> float_repr:
    sign = jnp.sign(x_in)
    x_in = jnp.abs(x_in)
    x_in = jax.lax.bitcast_convert_type(x_in, bits_dtype)
    if stochastic_rounding:
      if test_noise_axis is not None:
        noise_axis_size = x_in.shape[test_noise_axis]
        sh = list((1,) * len(x_in.shape))
        sh[test_noise_axis] = noise_axis_size
        rnd_bits = jnp.arange(noise_axis_size, dtype=bits_dtype).reshape(sh)
      else:
        rnd_bits = jax.random.bits(key, x_in.shape, bits_dtype)
      noise = jax.lax.bitwise_and(rnd_bits, mant_trunc_mask)
      x_in = x_in + noise
    else:
      carry_bit = bits_dtype(1 << (mant_trunc_bits - 1))
      x_in = x_in + carry_bit
    x_in = jax.lax.bitwise_and(x_in, ~mant_trunc_mask)
    x_in = jax.lax.bitcast_convert_type(x_in, input_dtype)
    if nexp is not None:
      x_in = jnp.clip(x_in, min_normal, max_normal)
    return x_in * sign

  def round_subnormal(x_in: float_repr) -> float_repr:
    if cfg.has_subnormals:
      if stochastic_rounding:
        if test_noise_axis is not None:
          noise_axis_size = x_in.shape[test_noise_axis]
          sh = list((1,) * len(x_in.shape))
          sh[test_noise_axis] = noise_axis_size
          rnd_bits = jnp.arange(noise_axis_size, dtype=bits_dtype).reshape(sh)
        else:
          rnd_bits = jax.random.bits(key, x.shape, bits_dtype)
        rnd_bits = (jnp.uint32(rnd_bits) << 7) | jnp.float32(1).view(jnp.uint32)
        noise = (
            jax.lax.bitcast_convert_type(rnd_bits, jnp.float32)
            - 1.5
            + 2 ** (-17)  # to make noise symmetric to zero
        )
      else:
        noise = 0.0
      x_in = (jnp.round(x_in * 2**nmant + noise) / 2**nmant).astype(input_dtype)
    else:
      if stochastic_rounding:
        if test_noise_axis is not None:
          noise_axis_size = x_in.shape[test_noise_axis]
          sh = list((1,) * len(x_in.shape))
          sh[test_noise_axis] = noise_axis_size
          rnd_bits = jnp.arange(noise_axis_size, dtype=bits_dtype).reshape(sh)
        else:
          rnd_bits = jax.random.bits(key, x_in.shape, bits_dtype)
        rnd_bits = (jnp.uint32(rnd_bits) << 7) | jnp.float32(2).view(jnp.uint32)
        rnd_floats = jax.lax.bitcast_convert_type(rnd_bits, jnp.float32) - (
            3.0 - 2 ** (-16)
        )
        x_in = (x_in > rnd_floats) * 2 - 1.0
      else:
        sign = jax.lax.bitwise_and(
            jax.lax.bitcast_convert_type(x_in, bits_dtype), sign_mask
        )
        x_in = jax.lax.bitwise_or(sign, min_normal_repr)
        x_in = jax.lax.bitcast_convert_type(x_in, input_dtype)
    return x_in

  normal = trunc_and_clip_normal(x)
  subnormal = round_subnormal(x)
  out = jnp.where(jnp.abs(x) < min_normal, subnormal, normal)
  # Deal with overflow
  if nexp == 0 and cfg.has_subnormals:
    out = jnp.clip(out, -max_subnormal, max_subnormal)
  return out


def fp_largest_representable(cfg: FpNumericsConfig):
  """Compute maximum representable FP number."""
  nexp = cfg.nexp
  nmant = cfg.nmant
  assert cfg.minexp == 0, 'minexp not implemented'
  assert not cfg.has_naninf, 'naninf not implemented'
  if cfg == RADIX4:
    return 64.0
  else:
    assert cfg.radix == 2, 'Only radix2 or radix4 is supported.'
  max_exp = 2**nexp - 1 + cfg.minexp - (1 if cfg.has_subnormals else 0)
  max_mant = 2**nmant - 1 - (1 if cfg.has_two_nan else 0)
  max_normal = 2**max_exp * (1 + (max_mant / 2**nmant))
  # TODO(yichizh): check max_subnormal again when minexp is not 0.
  max_subnormal = 2**cfg.minexp * max_mant / 2**nmant
  if cfg.has_subnormals and nexp == 0:
    # When there is no exponent bits, e.g., e0mX, compute the maximum
    # representable value by max subnormal instead of max normal.
    return max_subnormal
  else:
    return max_normal


@utils.flax_slots_kw_only_dataclass
class FpNumerics(numerics.AqtNumerics):
  """Numerics for fp8."""

  # Requested rounding precision.
  cfg: FpNumericsConfig = utils.static_field()
  stochastic_rounding: bool = utils.static_field(default=False)
  clip_gradient: bool = utils.static_field(default=False)

  def get_quant_bound(self):
    return fp_largest_representable(cfg=self.cfg)

  def get_dtype(self):
    return jnp.bfloat16

  def vjp_fwd(self, x, context):
    x = fp_round(
        x.astype(jnp.bfloat16),
        cfg=self.cfg,
        key=context.key,
        stochastic_rounding=self.stochastic_rounding,
    )
    res = (x,)
    return x, res

  def vjp_bwd(self, res, grad):
    ret = grad
    if self.clip_gradient:
      (x,) = res
      clip_bound = self.get_quant_bound()
      ret *= (-clip_bound <= x) * (x <= clip_bound)
    return (ret, None)


def radix4_round(
    x,
    *,
    cfg: FpNumericsConfig,
    key: jax.Array,
    stochastic_rounding: bool,
    test_noise_axis=None,
    coeff: float = 1.6,
):
  """Round to nearest radix4 (4^n)."""
  # TODO(yichizh): are the bucket centers the same in sc=True/False?
  assert cfg == RADIX4
  input_dtype = x.dtype
  msg = f'Unsupported dtype for sthochastic rounding: {input_dtype}.'
  bits_dtype = jnp.uint16 if input_dtype == jnp.bfloat16 else jnp.uint32
  assert input_dtype == jnp.bfloat16, msg
  x = x.astype(jnp.float32)  # upcast for log4 computation
  if stochastic_rounding:
    abs_x = jnp.abs(x)
    lower_bucket_ebit = jnp.floor(jnp.log(abs_x) / jnp.log(4))
    lower_bucket_center = 4**lower_bucket_ebit
    higher_bucket_center = 4 ** (lower_bucket_ebit + 1)
    # When input is 0, lower and higher bucket center are both -inf.
    # Hence bucket size is 0.
    bucket_size = higher_bucket_center - lower_bucket_center
    if test_noise_axis is not None:
      noise_axis_size = x.shape[test_noise_axis]
      msg = 'Input repetition should be consistent with effective noise bits.'
      assert noise_axis_size == 2**16, msg
      sh = list((1,) * len(x.shape))
      sh[test_noise_axis] = noise_axis_size
      rnd_bits = jnp.arange(noise_axis_size, dtype=bits_dtype).reshape(sh)
    else:
      rnd_bits = jax.random.bits(key, x.shape, bits_dtype)
    rnd_bits = (jnp.uint32(rnd_bits) << 7) | jnp.float32(1).view(jnp.uint32)
    noise = jax.lax.bitcast_convert_type(rnd_bits, jnp.float32) - 1.0  # [0, 1)
    noise = noise * bucket_size
    ebit = jnp.floor(jnp.log(abs_x + noise) / jnp.log(4))
  else:
    ebit = jnp.floor(jnp.log(jnp.abs(coeff * x)) / jnp.log(4))
  subnormal = jnp.zeros_like(x)
  overflow = jnp.sign(x) * 64
  radix4 = jnp.sign(x) * 4**ebit
  ret = jnp.where(ebit < -3, subnormal, jnp.where(ebit >= 3, overflow, radix4))
  return ret.astype(input_dtype)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test for pallas_tensor."""
from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2.pallas import pallas_tensor
from jax.experimental import pallas as pl
import jax.numpy as jnp


QTensor = aqt_tensor.QTensor
TransposedTensor = pallas_tensor.TransposedTensor


class PallasTensorTest(parameterized.TestCase):

  @parameterized.named_parameters(
      dict(
          testcase_name="2d_tensor_channelwise_1",
          qvalue_shape=(512, 512),
          scale_shape=(512, 1),
          block_shape=(128, 128),
          expected_scale_block_shape=(128, 1),
          index_and_expected_index=((2, 3), (2, 0)),
      ),
      dict(
          testcase_name="2d_tensor_channelwise_2",
          qvalue_shape=(512, 512),
          scale_shape=(1, 512),
          block_shape=(128, 128),
          expected_scale_block_shape=(1, 128),
          index_and_expected_index=((2, 3), (0, 3)),
      ),
      dict(
          testcase_name="3d_tensor_channelwise_1",
          qvalue_shape=(10, 512, 512),
          scale_shape=(10, 512, 1),
          block_shape=(1, 128, 128),
          expected_scale_block_shape=(1, 128, 1),
          index_and_expected_index=((3, 2, 3), (3, 2, 0)),
      ),
      dict(
          testcase_name="3d_tensor_channelwise_2",
          qvalue_shape=(10, 512, 512),
          scale_shape=(10, 1, 512),
          block_shape=(1, 128, 128),
          expected_scale_block_shape=(1, 1, 128),
          index_and_expected_index=((3, 2, 3), (3, 0, 3)),
      ),
  )
  def test_qtensor_blockspec_correctness(
      self,
      qvalue_shape,
      scale_shape,
      block_shape,
      expected_scale_block_shape,
      index_and_expected_index,
  ):
    qt = QTensor(
        qvalue=jnp.ones(qvalue_shape, dtype=jnp.int8),
        scale=[jnp.ones(scale_shape, dtype=jnp.float32)],
        scale_t=None,
        bias=[],
        dequant_dtype=jnp.float32,
    )
    block_spec = pl.BlockSpec(block_shape, lambda *args: args)
    qt_block_spec = pallas_tensor.make_qtensor_blockspec(qt, block_spec)

    self.assertEqual(qt_block_spec.qvalue, block_spec)
    self.assertEqual(
        qt_block_spec.scale[0].block_shape, expected_scale_block_shape
    )

    index, expected_index = index_and_expected_index
    self.assertEqual(qt_block_spec.scale[0].index_map(*index), expected_index)

  @parameterized.named_parameters(
      dict(
          testcase_name="2d_tensor_channelwise_1",
          tensor_shape=(512, 1),
          block_shape=(128, 1),
          expect_transpose=True,
          expected_permute_axes=[1, 0],
          expected_transposed_tensor_shape=(1, 512),
          expected_block_shape=(1, 128),
      ),
      dict(
          testcase_name="2d_tensor_channelwise_2",
          tensor_shape=(1, 512),
          block_shape=(1, 128),
          expect_transpose=False,
      ),
      dict(
          testcase_name="3d_tensor_channelwise_1",
          tensor_shape=(10, 512, 1),
          block_shape=(1, 128, 1),
          expect_transpose=True,
          expected_permute_axes=[0, 2, 1],
          expected_transposed_tensor_shape=(10, 1, 512),
          expected_block_shape=(1, 1, 128),
      ),
      dict(
          testcase_name="3d_tensor_channelwise_2",
          tensor_shape=(10, 1, 512),
          block_shape=(1, 1, 128),
          expect_transpose=False,
      ),
  )
  def test_transpose_for_memory_saving(
      self,
      tensor_shape,
      block_shape,
      expect_transpose,
      expected_permute_axes=None,
      expected_transposed_tensor_shape=None,
      expected_block_shape=None,
  ):
    t = jnp.ones(tensor_shape)
    block_spec = pl.BlockSpec(block_shape, lambda *args: args)

    transposed_t, transposed_block_spec = (
        pallas_tensor.transpose_tensor_for_memory_saving(t, block_spec)
    )

    if expect_transpose:
      self.assertIsInstance(transposed_t, TransposedTensor)
      self.assertIsInstance(transposed_block_spec, TransposedTensor)
      self.assertEqual(transposed_t.permute_axes, expected_permute_axes)
      self.assertEqual(
          transposed_t.transposed_tensor.shape, expected_transposed_tensor_shape
      )
      self.assertEqual(
          transposed_block_spec.transposed_tensor.block_shape,
          expected_block_shape,
      )
    else:
      self.assertTrue((transposed_t == t).all())
      self.assertEqual(transposed_block_spec, block_spec)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test for dot_general for pallas."""

import functools
import math
from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import aqt_tensor
import aqt.jax.v2.pallas as aqt_pl
import jax
from jax.experimental import pallas as pl
import jax.numpy as jnp


QTensor = aqt_tensor.QTensor
DequantMode = aqt_pl.DequantMode


class DotGeneralTest(parameterized.TestCase):

  @parameterized.product(
      mkn_and_blk=[
          ((512, 512, 512), (128, 128, 128)),
          ((1024, 1024, 1024), (256, 128, 256)),
          ((2048, 2048, 2048), (512, 512, 512)),
          ((4096, 4096, 4096), (512, 512, 512)),
          ((8192, 8192, 8192), (512, 512, 512)),
      ],
      quantize_lhs=[True, False],
      quantize_rhs=[True, False],
  )
  def test_quantized_matmul_error(
      self, mkn_and_blk, quantize_lhs, quantize_rhs
  ):
    (m, k, n), mkn_blk = mkn_and_blk

    key = jax.random.PRNGKey(0)
    lhs = jax.random.uniform(key, shape=(m, k))
    rhs = jax.random.uniform(key, shape=(k, n))

    @functools.partial(jax.jit, static_argnames=["block_size"])
    def q_matmul(lhs, rhs, block_size=(512, 512, 512)):
      assert lhs.shape[1] == rhs.shape[0]
      m, k = lhs.shape
      _, n = rhs.shape

      m_blk, k_blk, n_blk = block_size

      if quantize_lhs:
        lhs = aqt_pl.quant(lhs, n_bits=8, calibration_axes=(1,))
      if quantize_rhs:
        rhs = aqt_pl.quant(rhs, n_bits=8, calibration_axes=(0,))

      def kernel(
          lhs_ref: QTensor,
          rhs_ref: QTensor,
          out_ref: jnp.ndarray,
      ):
        @pl.when(pl.program_id(axis=2) == 0)
        def _():
          out_ref[...] = jnp.zeros_like(out_ref)

        if isinstance(lhs_ref, QTensor):
          lhs = aqt_pl.load_qtensor(lhs_ref)
        else:
          lhs = lhs_ref[...]

        if isinstance(rhs_ref, QTensor):
          rhs = aqt_pl.load_qtensor(rhs_ref)
        else:
          rhs = rhs_ref[...]

        out = aqt_pl.dot_general(
            lhs,
            rhs,
            dimension_numbers=(((1,), (0,)), ((), ())),
        )
        out_ref[...] += out

      lhs_inspec = pl.BlockSpec((m_blk, k_blk), lambda i, j, k: (i, k))
      rhs_inspec = pl.BlockSpec((k_blk, n_blk), lambda i, j, k: (k, j))
      out_spec = pl.BlockSpec((m_blk, n_blk), lambda i, j, k: (i, j))

      out = aqt_pl.pallas_call(
          kernel,
          out_shape=jax.ShapeDtypeStruct((m, n), jnp.float32),
          grid=(
              math.ceil(m / m_blk),
              math.ceil(n // n_blk),
              math.ceil(k // k_blk),
          ),
          in_specs=[
              lhs_inspec,
              rhs_inspec,
          ],
          out_specs=out_spec,
          interpret=False,
      )(lhs, rhs)
      return out

    quantized_out = q_matmul(lhs, rhs, block_size=mkn_blk)
    reference_out = lhs @ rhs

    difference = jnp.abs(quantized_out - reference_out)
    relative_error = difference / jnp.abs(reference_out)
    max_relative_error = relative_error.max().item()

    if quantize_lhs or quantize_rhs:
      self.assertLess(max_relative_error, 1e-2)
    else:
      # If none of tensor are quantized, the result should exactly identical
      # or the error should be very small.
      self.assertLess(max_relative_error, 1e-6)

  @parameterized.named_parameters(
      dict(
          testcase_name="OUTPUT_lhs_OUTPUT_rhs",
          lhs_calibration_axes=(1,),
          rhs_calibration_axes=(0,),
          dequant_mode_lhs=DequantMode.OUTPUT,
          dequant_mode_rhs=DequantMode.OUTPUT,
      ),
      dict(
          testcase_name="THIS_INPUT_lhs_OUTPUT_rhs",
          lhs_calibration_axes=(1,),
          rhs_calibration_axes=(0,),
          dequant_mode_lhs=DequantMode.THIS_INPUT,
          dequant_mode_rhs=DequantMode.OUTPUT,
      ),
      dict(
          testcase_name="OUTPUT_lhs_THIS_INPUT_rhs",
          lhs_calibration_axes=(1,),
          rhs_calibration_axes=(0,),
          dequant_mode_lhs=DequantMode.OUTPUT,
          dequant_mode_rhs=DequantMode.THIS_INPUT,
      ),
      dict(
          testcase_name="OTHER_INPUT_lhs",
          lhs_calibration_axes=(0,),
          rhs_calibration_axes=None,
          dequant_mode_lhs=DequantMode.OTHER_INPUT,
      ),
      dict(
          testcase_name="OTHER_INPUT_rhs",
          lhs_calibration_axes=None,
          rhs_calibration_axes=(1,),
          dequant_mode_rhs=DequantMode.OTHER_INPUT,
      ),
  )
  def test_dequantization_location(
      self,
      lhs_shape=(2048, 512),
      rhs_shape=(512, 2048),
      lhs_calibration_axes=None,
      rhs_calibration_axes=None,
      dequant_mode_lhs=DequantMode.OUTPUT,
      dequant_mode_rhs=DequantMode.OUTPUT,
      dimension_numbers=(((1,), (0,)), ((), ())),
  ):
    lhs = jnp.zeros(lhs_shape)
    rhs = jnp.zeros(rhs_shape)

    if lhs_calibration_axes is not None:
      lhs_qtensor = aqt_pl.quant(
          lhs, n_bits=8, calibration_axes=lhs_calibration_axes
      )
    else:
      lhs_qtensor = lhs

    if rhs_calibration_axes is not None:
      rhs_qtensor = aqt_pl.quant(
          rhs, n_bits=8, calibration_axes=rhs_calibration_axes
      )
    else:
      rhs_qtensor = rhs

    # Calling this function implictly pass the tests.
    aqt_pl.dot_general(
        lhs_qtensor,
        rhs_qtensor,
        dimension_numbers=dimension_numbers,
        lhs_dequant_mode=dequant_mode_lhs,
        rhs_dequant_mode=dequant_mode_rhs,
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""dot_general with flax lifted custom_vjp."""
from aqt.jax.v2 import aqt_dot_general
from aqt.jax.v2 import aqt_tensor
import flax.linen as nn
import jax
from jax import numpy as jnp


def dg_core_flax_lifted(
    lhs: jnp.ndarray,
    rhs: jnp.ndarray,
    lhs_qt: None | aqt_tensor.QTensor,
    rhs_qt: None | aqt_tensor.QTensor,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    mdl: nn.Module,
    cfg: aqt_dot_general.DotGeneral,
):
  """dot_general with flax lifted custom_vjp applied on it.

  Args:
    lhs: Left hand side argument of dot_general.
    rhs: Right hand side argument of dot_general.
    lhs_qt: Left hand side QTensor, if applicable.
    rhs_qt: Right hand side QTensor, if applicable.
    dimension_numbers: Dimension numbers for dot_general.
    mdl: Flax module in which the dot_general is called.
    cfg: aqt DotGeneral.

  Returns:
    aqt DotGeneral result. Flax-lifted custom_vjp is applied on it.
  """

  def _dg_core_flax_lifted(
      mdl: nn.Module,
      lhs: jnp.ndarray,
      rhs: jnp.ndarray,
      lhs_qt: None | aqt_tensor.QTensor,
      rhs_qt: None | aqt_tensor.QTensor,
      dimension_numbers: jax.lax.DotDimensionNumbers,
      cfg: aqt_dot_general.DotGeneral,
  ):
    """Lifted dg_core."""
    # The order of parameters must match with that of jax version to avoid the
    # backward gradient - primal order mismatch.
    del mdl
    return cfg.dg_core(
        lhs=lhs,
        rhs=rhs,
        lhs_qt=lhs_qt,
        rhs_qt=rhs_qt,
        dimension_numbers=dimension_numbers,
    )

  def _dg_core_vjp_fwd_flax_lifted(
      mdl: nn.Module,
      lhs: jnp.ndarray,
      rhs: jnp.ndarray,
      lhs_qt: None | aqt_tensor.QTensor,
      rhs_qt: None | aqt_tensor.QTensor,
      dimension_numbers: jax.lax.DotDimensionNumbers,
      cfg: aqt_dot_general.DotGeneral,
  ):
    """Lifted custom vjp_fwd."""

    # Currently we do not support backward computation for the variables
    # declared INSIDE the AqtDotGeneral layer.
    # Since we do not have any gradient functions for the params, removing
    # jax.lax.stop_gradient here will lead to NotImplementedError of
    # differentiation rules for 'custom_lin'. MutableArrays should not be passed
    # through stop_gradient, so we need to filter those out from this logic.
    def stop_grad_for_non_ref_params(param):
      if isinstance(
          jax.core.get_aval(param),
          # pylint: disable-next=protected-access
          jax._src.state.types.AbstractRef,
      ):
        return param
      return jax.lax.stop_gradient(param)

    params = jax.tree_util.tree_map(stop_grad_for_non_ref_params, mdl.variables)
    out, res = aqt_dot_general.dg_core_vjp_fwd(
        lhs, rhs, lhs_qt, rhs_qt, dimension_numbers, cfg
    )
    return out, (res, params)

  def _dg_core_vjp_bwd_flax_lifted(
      fwd_dimension_numbers: jax.lax.DotDimensionNumbers,
      res: tuple[
          tuple[
              None | aqt_dot_general.DotGeneralRes,
              aqt_dot_general.DotGeneral,
          ],
          jax.core.ParamDict,
      ],
      g,
  ):
    """Lifted custom vjp_bwd."""
    res, params = res
    tangents = aqt_dot_general.dg_core_vjp_bwd(fwd_dimension_numbers, res, g)

    return params, *tangents

  dg_core_with_custom_vjp = nn.custom_vjp(
      _dg_core_flax_lifted,
      _dg_core_vjp_fwd_flax_lifted,
      _dg_core_vjp_bwd_flax_lifted,
      # Having a variable collection which does not included in the fwd/
      # bwd pipeline would trigger a complex pytype tree matching issue.
      grad_vars=True,
      nondiff_argnums=(5,),
  )
  return dg_core_with_custom_vjp(
      mdl, lhs, rhs, lhs_qt, rhs_qt, dimension_numbers, cfg
  )



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This test file only tests utility functions for gptq dot_general.

For e2e test, see flax_e2e_model_test.py.
"""

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2.extensions.gptq import gptq_dot_general_quantizer
import jax


class GptqDotGeneralQuantizerTest(parameterized.TestCase):

  @parameterized.parameters([
      dict(
          param_shape=(4, 5, 16),
          ca=[2],
          block_size=8,
          expected_converted_kernel_shape=(2, 8, 20),
      ),
      dict(
          param_shape=(4, 5, 6, 8),
          ca=[1, 3],
          block_size=4,
          expected_converted_kernel_shape=(10, 4, 24),
      ),
      dict(
          param_shape=(6, 8, 10, 12),
          ca=[0, 1, 2],
          block_size=24,
          expected_converted_kernel_shape=(20, 24, 12),
      ),
  ])
  def test_kernel_shape_transform(
      self, param_shape, ca, block_size, expected_converted_kernel_shape
  ):
    """Tests kernel transform before and after applying GPTQ."""

    kernel = jax.random.uniform(jax.random.PRNGKey(0), shape=param_shape)
    new_kernel, kernel_feature_grouped_shape = (
        gptq_dot_general_quantizer._reshape_kernel_for_gptq(
            kernel, ca, None, False, None, block_size
        )
    )
    self.assertEqual(expected_converted_kernel_shape, new_kernel.shape)
    recovered_kernel = (
        gptq_dot_general_quantizer._recover_kernel_from_gptq_result(
            new_kernel,
            ca,
            None,
            False,
            None,
            kernel.dtype,
            kernel_feature_grouped_shape,
        )
    )

    assert (kernel == recovered_kernel).all()


if __name__ == "__main__":
  absltest.main()
